# -*- coding: utf-8 -*-
"""
Test the base flow class
"""
import pytest
import torch
from unittest.mock import MagicMock, create_autospec, patch

from nessai.flows.base import BaseFlow, NFlow
from glasflow.nflows.transforms import Transform
from glasflow.nflows.distributions import Distribution


@pytest.fixture
def flow():
    return create_autospec(BaseFlow)


@pytest.fixture
def nflow():
    nflow = create_autospec(NFlow)
    nflow._transform = MagicMock(spec=Transform)
    nflow._distribution = MagicMock(spec=Distribution)
    return nflow


def test_base_flow_abstract_methods():
    """Assert an error is raised if the methods are not implemented."""
    with pytest.raises(TypeError) as excinfo:
        BaseFlow()
    assert "instantiate abstract class BaseFlow" in str(excinfo.value)


@pytest.mark.parametrize(
    "method",
    [
        "forward",
        "inverse",
        "sample",
        "log_prob",
        "base_distribution_log_prob",
        "forward_and_log_prob",
        "sample_and_log_prob",
        "sample_latent_distribution",
    ],
)
def test_base_flow_methods(method, flow):
    """Test to make sure all of the method in the base class raise errors"""
    with pytest.raises(NotImplementedError):
        getattr(BaseFlow, method)(flow, None)


@pytest.mark.parametrize(
    "method",
    [
        "freeze_transform",
        "unfreeze_transform",
    ],
)
def test_base_flow_methods_no_args(method, flow):
    """Assert methods with args and are not implemented raise an error"""
    with pytest.raises(NotImplementedError):
        getattr(BaseFlow, method)(flow)


@pytest.mark.parametrize("method", ["end_iteration", "finalise"])
def test_base_flow_pass_methods(method, flow):
    """Test the methods that don't do anything in the base class."""
    assert getattr(BaseFlow, method)(flow) is None


def test_base_flow_to(flow):
    """Assert the to method works as intended"""
    flow.device = None
    with patch("nessai.flows.base.Module.to") as super_init:
        BaseFlow.to(flow, "cpu")
    super_init.assert_called_once_with("cpu")
    assert flow.device == "cpu"


def test_nflow_base_transform():
    """Assert an error is raised if the transform class is invalid."""

    class Test:
        pass

    with pytest.raises(TypeError) as excinfo:
        NFlow(Test(), Test())
    assert "transform must inherit" in str(excinfo.value)


def test_nflow_base_distribution():
    """Assert an error is raised if the distribution class is invalid."""

    class Test:
        pass

    with pytest.raises(TypeError) as excinfo:
        NFlow(Transform(), Test())
    assert "distribution must inherit" in str(excinfo.value)


def test_nflow_finalise(nflow):
    """Assert the methods are called"""
    nflow._transform.finalise = MagicMock()
    nflow._distribution.finalise = MagicMock()
    NFlow.finalise(nflow)
    nflow._transform.finalise.assert_called_once()
    nflow._distribution.finalise.assert_called_once()


def test_nflow_finalise_not_called(nflow):
    """Assert no error is raised the methods are missing"""
    NFlow.finalise(nflow)


def test_nflow_end_iteration(nflow):
    """Assert the methods are called"""
    nflow._transform.end_iteration = MagicMock()
    nflow._distribution.end_iteration = MagicMock()
    NFlow.end_iteration(nflow)
    nflow._transform.end_iteration.assert_called_once()
    nflow._distribution.end_iteration.assert_called_once()


def test_nflow_end_iteration_not_called(nflow):
    """Assert no error is raised the methods are missing"""
    NFlow.end_iteration(nflow)


def test_nflow_freeze(nflow):
    """Assert the transform is frozen"""
    nflow._transform.requires_grad_ = MagicMock()
    NFlow.freeze_transform(nflow)
    nflow._transform.requires_grad_.assert_called_once_with(False)


def test_nflow_unfreeze(nflow):
    """Assert the transform is unfrozen"""
    nflow._transform.requires_grad_ = MagicMock()
    NFlow.unfreeze_transform(nflow)
    nflow._transform.requires_grad_.assert_called_once_with(True)


def test_nflow_sample_latent_distribution(nflow):
    n = 10
    NFlow.sample_latent_distribution(nflow, n)
    nflow._distribution.sample.assert_called_once_with(n)


def test_nflow_sample_latent_distribution_context(nflow):
    n = 10
    context = torch.randn(10)
    with pytest.raises(NotImplementedError):
        NFlow.sample_latent_distribution(nflow, n, context=context)



# -*- coding: utf-8 -*-
"""
Tests related to checking and updating the state of sampler and the history.
"""
import os
import pytest
from unittest.mock import MagicMock, patch

from nessai.livepoint import parameters_to_live_point
from nessai.samplers.nestedsampler import NestedSampler


@pytest.fixture
def sampler(sampler):
    """A sampler mock configured to work with update state"""
    # Stored stats
    sampler.history = dict(
        iterations=[10],
        population_radii=[1.0],
        population_iterations=[0],
        population_acceptance=[0.5],
        likelihood_evaluations=[10],
        min_log_likelihood=[0.0],
        max_log_likelihood=[100.0],
        logZ=[-100.0],
        dlogZ=[100.0],
        mean_acceptance=[0.5],
    )
    # Attributed used to update stats
    sampler.model = MagicMock()
    sampler.model.likelihood_evaluations = 100
    sampler.logLmin = 0.0
    sampler.logLmax = 150.0
    sampler.condition = 50.0

    sampler.state = MagicMock()
    sampler.state.logZ = -50.0
    sampler.state.info = [0.0]
    sampler.state.log_evidence_error = 0.1

    sampler.proposal = MagicMock()
    sampler.proposal.r = 2.0
    sampler.proposal.population_acceptance = 0.4

    sampler.mean_acceptance = 0.5
    sampler.mean_block_acceptance = 0.5
    sampler.block_acceptance = 0.5
    sampler.block_iteration = 5

    sampler.checkpoint = MagicMock()

    return sampler


@pytest.mark.parametrize("switch", [False, True])
@pytest.mark.parametrize("uninformed", [False, True])
def test_check_state_force(sampler, switch, uninformed):
    """Test the behaviour of check_state with force=True.

    Training should always start irrespective of other checks and with
    force=True unless uninformed sampling is being used and the switch=False.
    """
    sampler.uninformed_sampling = uninformed
    sampler.check_proposal_switch = MagicMock(return_value=switch)
    sampler.check_training = MagicMock()
    sampler.train_proposal = MagicMock()

    NestedSampler.check_state(sampler, force=True)

    if uninformed and not switch:
        sampler.train_proposal.assert_not_called()
    else:
        sampler.train_proposal.assert_called_once_with(force=True)

    sampler.check_training.assert_not_called()


@pytest.mark.parametrize("force", [False, True])
@pytest.mark.parametrize("train", [False, True])
def test_check_state_train(sampler, force, train):
    """Test the behaviour of check_state with force=False and `check_training`
    returns True, False, True, True, False, True, or False, False.

    Force is used for the return values for `check_training`
    """
    sampler.uninformed_sampling = False
    sampler.check_proposal_switch = MagicMock()
    sampler.check_training = MagicMock(return_value=(train, force))
    sampler.train_proposal = MagicMock()

    NestedSampler.check_state(sampler, force=False)
    if train or force:
        sampler.check_training.assert_called_once_with()
        sampler.train_proposal.assert_called_once_with(force=force)
    else:
        sampler.check_training.assert_called_once_with()
        sampler.train_proposal.assert_not_called()


def test_update_history(sampler):
    """Test updating the history dictionary"""

    sampler.iteration = 15

    with patch(
        "nessai.samplers.nestedsampler.BaseNestedSampler.update_history"
    ) as mock:
        NestedSampler.update_history(sampler)

    mock.assert_called_once()

    assert sampler.history["population_acceptance"] == [0.5]
    assert sampler.history["min_log_likelihood"] == [0.0, 0.0]
    assert sampler.history["max_log_likelihood"] == [100.0, 150.0]
    assert sampler.history["logZ"] == [-100.0, -50.0]
    assert sampler.history["dlogZ"] == [100.0, 50.0]
    assert sampler.history["mean_acceptance"] == [0.5, 0.5]
    assert sampler.history["iterations"] == [10, 15]
    sampler.checkpoint.assert_not_called()


@patch("nessai.samplers.nestedsampler.NestedSampler.checkpoint")
def test_update_state_checked_acceptance(mock, sampler):
    """Test the behaviour of update state if `_checked_population` is False.

    Checks to make sure the correct statistics are stored and the other
    checks are not called because it / nlive and it / (nlive/10) != 0
    """
    sampler.iteration = 11
    sampler.proposal._checked_population = False
    sampler.checkpointing = False
    sampler.update_history = MagicMock()

    NestedSampler.update_state(sampler)

    sampler.update_history.assert_not_called()

    assert sampler.history["population_acceptance"] == [0.5, 0.4]
    assert sampler.history["population_radii"] == [1.0, 2.0]
    assert sampler.history["population_iterations"] == [0, 11]
    assert sampler.proposal._checked_population is True
    mock.assert_not_called()


def test_update_state_history(sampler):
    """Test the behaviour of updated state if it / (nlive/10) == 0

    Checks to make sure the correct statistics are saved and metrics related to
    population are not updated.
    """
    sampler.iteration = 10
    sampler.proposal._checked_population = True
    sampler.checkpointing = False
    sampler.update_history = MagicMock()

    NestedSampler.update_state(sampler)

    sampler.update_history.assert_called_once()

    assert sampler.history["population_acceptance"] == [0.5]
    sampler.checkpoint.assert_not_called()

    assert sampler.proposal.ns_acceptance == 0.5


@pytest.mark.parametrize("plot", [False, True])
@patch("nessai.samplers.nestedsampler.plot_indices")
def test_update_state_every_nlive(mock_plot, plot, sampler):
    """Test the update that happens every nlive iterations.

    Tests both with plot=True and plot=False
    """
    sampler.nlive = 100
    sampler.iteration = 100
    sampler.proposal._checked_population = True
    sampler.check_insertion_indices = MagicMock()
    sampler.plot = plot
    sampler.uninformed_sampling = True
    sampler.plot_state = MagicMock()
    sampler.plot_trace = MagicMock()
    sampler.output = os.getcwd()
    sampler.insertion_indices = range(2 * sampler.nlive)
    sampler.checkpointing = False
    sampler.update_history = MagicMock()

    NestedSampler.update_state(sampler)

    sampler.update_history.assert_called()
    sampler.check_insertion_indices.assert_called_once()
    assert sampler.block_iteration == 0
    assert sampler.block_acceptance == 0.0
    assert sampler.history["population_acceptance"] == [0.5]

    if plot:
        sampler.plot_state.assert_called_once_with(
            filename=os.path.join(os.getcwd(), "state.png")
        )
        sampler.plot_trace.assert_called_once()
        mock_plot.assert_called_once_with(
            sampler.insertion_indices[-100:],
            100,
            plot_breakdown=False,
            filename=os.path.join(
                os.getcwd(), "diagnostics", "insertion_indices_100.png"
            ),
        )
    else:
        sampler.plot_state.assert_not_called()
        sampler.plot_trace.assert_not_called()
        mock_plot.assert_not_called()


@patch("nessai.samplers.nestedsampler.plot_indices")
def test_update_state_force(mock_plot, sampler):
    """Test the update that happens if force=True.

    Checks that plot_indices is not called even if plotting is enabled.
    """
    sampler.iteration = 111
    sampler.proposal._checked_population = True
    sampler.check_insertion_indices = MagicMock()
    sampler.plot = True
    sampler.plot_state = MagicMock()
    sampler.plot_trace = MagicMock()
    sampler.output = os.getcwd()
    sampler.uninformed_sampling = False
    sampler.checkpointing = False
    sampler.update_history = MagicMock()

    NestedSampler.update_state(sampler, force=True)

    sampler.update_history.assert_called_once()

    mock_plot.assert_not_called()
    sampler.plot_trace.assert_called_once()
    sampler.plot_state.assert_called_once_with(
        filename=os.path.join(os.getcwd(), "state.png")
    )

    assert sampler.history["population_acceptance"] == [0.5]
    assert sampler.block_acceptance == 0.5
    assert sampler.block_iteration == 5


def test_update_state_checkpointing(sampler):
    """Assert the checkpoint function is called"""
    sampler.checkpointing = True
    sampler.checkpoint = MagicMock()
    sampler.iteration = 10
    NestedSampler.update_state(sampler)
    sampler.checkpoint.assert_called_once_with(periodic=True)


def test_update_state_checkpointing_disabled(sampler):
    """Assert the checkpoint function is not called"""
    sampler.checkpointing = False
    sampler.checkpoint = MagicMock()
    sampler.iteration = 10
    NestedSampler.update_state(sampler)
    sampler.checkpoint.assert_not_called()


def test_get_result_dictionary(sampler):
    """Assert the correct dictionary is returned"""
    from datetime import timedelta

    base_result = dict(
        seed=1234,
        history=dict(
            likelihood_evaluations=[10],
            sampling_time=[2],
        ),
    )
    sampler.nlive = 1
    sampler.iteration = 3
    sampler.min_log_likelihood = [-3, -2, 1]
    sampler.max_log_likelihood = [1, 2, 3]
    sampler.likelihood_evaluations = 3
    sampler.logZ_history = [1, 2, 3]
    sampler.mean_acceptance_history = [1, 2, 3]
    sampler.rolling_p = [0.5]
    sampler.population_iterations = []
    sampler.population_acceptance = []
    sampler.training_iterations = []
    sampler.insertion_indices = []
    sampler.training_time = timedelta()
    sampler.proposal_population_time = timedelta()
    sampler.nested_samples = [
        parameters_to_live_point((1, 2), ["x", "y"]),
        parameters_to_live_point((3, 4), ["x", "y"]),
    ]
    sampler.final_p_value = 0.5
    sampler.final_ks_statistic = 0.1
    sampler.state.log_posterior_weights = [0.5, 0.5]

    with patch(
        "nessai.samplers.base.BaseNestedSampler.get_result_dictionary",
        return_value=base_result,
    ) as mock:
        out = NestedSampler.get_result_dictionary(sampler)

    mock.assert_called_once()

    assert out["seed"] == 1234
    assert "history" in out



# -*- coding: utf-8 -*-
"""
Integration tests for running the sampler with different configurations.
"""
import logging
import os
from scipy.stats import norm
import torch
import pytest
import numpy as np
from unittest.mock import patch

from nessai.flowsampler import FlowSampler
from nessai.livepoint import numpy_array_to_live_points
from nessai.model import Model
from nessai.utils.testing import IntegrationTestModel


torch.set_num_threads(1)


@pytest.mark.slow_integration_test
def test_sampling_with_rescale(integration_model, flow_config, tmpdir):
    """
    Test sampling with rescaling. Checks that flow is trained.
    """
    output = str(tmpdir.mkdir("w_rescale"))
    fp = FlowSampler(
        integration_model,
        output=output,
        resume=False,
        nlive=100,
        plot=False,
        flow_config=flow_config,
        training_frequency=10,
        maximum_uninformed=9,
        rescale_parameters=True,
        seed=1234,
        max_iteration=11,
        poolsize=10,
    )
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


@pytest.mark.slow_integration_test
def test_sampling_with_inversion(integration_model, flow_config, tmpdir):
    """
    Test sampling with inversion. Checks that flow is trained.
    """
    output = str(tmpdir.mkdir("w_rescale"))
    fp = FlowSampler(
        integration_model,
        output=output,
        resume=False,
        nlive=100,
        plot=False,
        flow_config=flow_config,
        training_frequency=10,
        maximum_uninformed=9,
        rescale_parameters=True,
        seed=1234,
        max_iteration=11,
        poolsize=10,
        boundary_inversion=True,
        update_bounds=True,
    )
    fp.run()
    reparams = list(fp.ns.proposal._reparameterisation.values())
    assert len(reparams) == 1
    assert reparams[0].parameters == ["x_0", "x_1"]
    assert list(reparams[0].boundary_inversion.keys()) == ["x_0", "x_1"]
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


def test_sampling_regex_reparams(model, flow_config, tmp_path):
    """Test using regex to specify reparameterisations"""
    model._names = ["x_0", "x_1"]
    model._bounds = {"x_0": [-5, 5], "x_1": [-5, 5]}

    fs = FlowSampler(
        model,
        nlive=100,
        output=tmp_path / "test_regex",
        flow_config=flow_config,
        reparameterisations={"z-score": {"parameters": ["x.*"]}},
        maximum_uninformed=50,
        max_iteration=100,
        plot=False,
    )
    fs.run()
    assert fs.ns._flow_proposal.rescaled_names == ["x_0_prime", "x_1_prime"]


@pytest.mark.slow_integration_test
def test_sampling_without_rescale(integration_model, flow_config, tmpdir):
    """
    Test sampling without rescaling. Checks that flow is trained.
    """
    output = str(tmpdir.mkdir("wo_rescale"))
    fp = FlowSampler(
        integration_model,
        output=output,
        resume=False,
        nlive=100,
        plot=False,
        flow_config=flow_config,
        training_frequency=10,
        maximum_uninformed=9,
        rescale_parameters=False,
        seed=1234,
        max_iteration=11,
        poolsize=10,
    )
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


@pytest.mark.slow_integration_test
def test_sampling_with_maf(integration_model, flow_config, tmpdir):
    """
    Test sampling with MAF. Checks that flow is trained but does not
    check convergence.
    """
    flow_config["model_config"]["ftype"] = "maf"
    output = str(tmpdir.mkdir("maf"))
    fp = FlowSampler(
        integration_model,
        output=output,
        resume=False,
        nlive=100,
        plot=False,
        flow_config=flow_config,
        training_frequency=10,
        maximum_uninformed=9,
        rescale_parameters=True,
        seed=1234,
        max_iteration=11,
        poolsize=10,
    )
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1


@pytest.mark.slow_integration_test
@pytest.mark.parametrize("analytic", [False, True])
def test_sampling_uninformed(integration_model, flow_config, tmpdir, analytic):
    """
    Test running the sampler with the two uninformed proposal methods.
    """
    output = str(tmpdir.mkdir("uninformed"))
    fp = FlowSampler(
        integration_model,
        output=output,
        resume=False,
        nlive=100,
        plot=False,
        flow_config=flow_config,
        training_frequency=None,
        maximum_uninformed=10,
        rescale_parameters=True,
        seed=1234,
        max_iteration=11,
        poolsize=10,
        analytic_priors=analytic,
    )
    fp.run()


@pytest.mark.slow_integration_test
@pytest.mark.parametrize("parallelise_prior", [True, False])
def test_sampling_with_n_pool(
    integration_model,
    flow_config,
    tmpdir,
    mp_context,
    parallelise_prior,
):
    """
    Test running the sampler with multiprocessing.
    """
    output = str(tmpdir.mkdir("pool"))
    with patch("multiprocessing.Pool", mp_context.Pool), patch(
        "nessai.utils.multiprocessing.multiprocessing.get_start_method",
        mp_context.get_start_method,
    ):
        fp = FlowSampler(
            integration_model,
            output=output,
            resume=False,
            nlive=100,
            plot=False,
            flow_config=flow_config,
            training_frequency=10,
            maximum_uninformed=9,
            rescale_parameters=True,
            seed=1234,
            max_iteration=11,
            poolsize=10,
            pytorch_threads=2,
            n_pool=2,
            parallelise_prior=parallelise_prior,
        )
    fp.run()
    assert fp.ns.proposal.flow.weights_file is not None
    assert fp.ns.proposal.training_count == 1
    assert os.path.exists(os.path.join(output, "result.hdf5"))


@pytest.mark.slow_integration_test
def test_sampling_resume(model, flow_config, tmpdir):
    """
    Test resuming the sampler.
    """
    output = str(tmpdir.mkdir("resume"))
    fp = FlowSampler(
        model,
        output=output,
        resume=True,
        nlive=100,
        plot=False,
        flow_config=flow_config,
        training_frequency=10,
        maximum_uninformed=9,
        rescale_parameters=True,
        checkpoint_on_iteration=True,
        checkpoint_interval=5,
        seed=1234,
        max_iteration=11,
        poolsize=10,
    )
    fp.run()
    assert os.path.exists(os.path.join(output, "nested_sampler_resume.pkl"))

    fp = FlowSampler(
        model,
        output=output,
        resume=True,
        flow_config=flow_config,
    )
    assert fp.ns.iteration == 11
    fp.ns.max_iteration = 21
    fp.run()
    assert fp.ns.iteration == 21
    assert os.path.exists(
        os.path.join(output, "nested_sampler_resume.pkl.old")
    )


@pytest.mark.slow_integration_test
def test_sampling_resume_w_pool(
    integration_model, flow_config, tmpdir, mp_context
):
    """
    Test resuming the sampler with a pool.
    """
    output = str(tmpdir.mkdir("resume"))
    with patch("multiprocessing.Pool", mp_context.Pool), patch(
        "nessai.utils.multiprocessing.multiprocessing.get_start_method",
        mp_context.get_start_method,
    ):
        fp = FlowSampler(
            integration_model,
            output=output,
            resume=True,
            nlive=100,
            plot=False,
            flow_config=flow_config,
            training_frequency=10,
            maximum_uninformed=9,
            rescale_parameters=True,
            checkpoint_on_iteration=True,
            checkpoint_interval=5,
            seed=1234,
            max_iteration=11,
            poolsize=10,
            n_pool=1,
        )
    assert fp.ns.model.n_pool == 1
    fp.run()
    assert os.path.exists(os.path.join(output, "nested_sampler_resume.pkl"))
    # Make sure the pool is already closed
    integration_model.close_pool()

    with patch("multiprocessing.Pool", mp_context.Pool), patch(
        "nessai.utils.multiprocessing.multiprocessing.get_start_method",
        mp_context.get_start_method,
    ):
        fp = FlowSampler(
            integration_model,
            output=output,
            resume=True,
            flow_config=flow_config,
            n_pool=1,
        )
    assert fp.ns.model.n_pool == 1
    assert fp.ns.iteration == 11
    fp.ns.max_iteration = 21
    fp.run()
    assert fp.ns.iteration == 21
    assert os.path.exists(
        os.path.join(output, "nested_sampler_resume.pkl.old")
    )


@pytest.mark.slow_integration_test
def test_sampling_resume_no_max_uninformed(
    integration_model, flow_config, tmpdir
):
    """
    Test resuming the sampler when there is no maximum iteration for
    the uinformed sampling.

    This test makes sure the correct proposal is loaded after resuming
    and re-initialising the sampler.
    """
    output = str(tmpdir.mkdir("resume"))
    fp = FlowSampler(
        integration_model,
        output=output,
        resume=True,
        nlive=100,
        plot=False,
        flow_config=flow_config,
        training_frequency=10,
        maximum_uninformed=9,
        rescale_parameters=True,
        seed=1234,
        max_iteration=11,
        checkpoint_on_iteration=True,
        checkpoint_interval=5,
        poolsize=10,
    )
    fp.run()
    assert os.path.exists(os.path.join(output, "nested_sampler_resume.pkl"))

    fp = FlowSampler(
        integration_model, output=output, resume=True, flow_config=flow_config
    )
    assert fp.ns.iteration == 11
    fp.ns.maximum_uninformed = np.inf
    fp.ns.initialise()
    assert fp.ns.proposal is fp.ns._flow_proposal
    fp.ns.max_iteration = 21
    fp.run()
    assert fp.ns.iteration == 21
    assert os.path.exists(
        os.path.join(output, "nested_sampler_resume.pkl.old")
    )


@pytest.mark.slow_integration_test
def test_resume_fallback_reparameterisation(tmpdir, model, flow_config):
    """
    Test resuming the sampler with the default reparameterisations disabled
    and the fallback set.

    This should load the same class after resuming.
    """
    from nessai.reparameterisations.rescale import ScaleAndShift

    output = str(tmpdir.mkdir("resume"))
    fp = FlowSampler(
        model,
        output=output,
        resume=True,
        nlive=100,
        plot=False,
        flow_config=flow_config,
        training_frequency=10,
        maximum_uninformed=9,
        rescale_parameters=False,
        use_default_reparameterisations=False,
        fallback_reparameterisation="z-score",
        checkpoint_on_iteration=True,
        checkpoint_interval=5,
        seed=1234,
        max_iteration=11,
        poolsize=10,
    )
    fp.run()

    reparam = fp.ns._flow_proposal._reparameterisation
    reparam = next(iter(fp.ns._flow_proposal._reparameterisation.values()))
    assert isinstance(reparam, ScaleAndShift)
    assert os.path.exists(os.path.join(output, "nested_sampler_resume.pkl"))

    fp = FlowSampler(
        model,
        output=output,
        resume=True,
        flow_config=flow_config,
    )
    assert fp.ns.iteration == 11
    fp.ns.max_iteration = 21
    fp.run()
    reparam = next(iter(fp.ns._flow_proposal._reparameterisation.values()))
    assert isinstance(reparam, ScaleAndShift)
    assert fp.ns.iteration == 21
    assert os.path.exists(
        os.path.join(output, "nested_sampler_resume.pkl.old")
    )


@pytest.mark.slow_integration_test
def test_sampling_with_infinite_prior_bounds(tmpdir):
    """
    Make sure the sampler runs when sampling a parameter with infinite prior \
        bounds.
    """
    output = str(tmpdir.mkdir("infinite_bounds"))

    class TestModel(Model):

        names = ["x", "y"]
        bounds = {"x": [0, 1], "y": [-np.inf, np.inf]}
        reparameterisations = {"x": "default", "y": None}

        def new_point(self, N=1):
            x = np.concatenate(
                [np.random.rand(N, 1), np.random.randn(N, 1)], axis=1
            )
            return numpy_array_to_live_points(x, self.names)

        def log_prior(self, x):
            log_p = np.log(self.in_bounds(x))
            log_p += norm.logpdf(x["y"])
            return log_p

        def log_likelihood(self, x):
            log_l = np.zeros(x.size)
            for n in self.names:
                log_l += norm.logpdf(x[n])
            return log_l

    fs = FlowSampler(
        TestModel(), output=output, nlive=500, plot=False, proposal_plots=False
    )
    fs.run(plot=False)
    assert fs.ns.condition <= 0.1


@pytest.mark.slow_integration_test
def test_constant_volume_mode(integration_model, tmpdir):
    """Test sampling in constant volume mode"""
    output = str(tmpdir.mkdir("test"))
    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=500,
        plot=False,
        proposal_plots=False,
        constant_volume_mode=True,
    )
    fs.run(plot=False)


@pytest.mark.slow_integration_test
def test_sampling_with_plotting(integration_model, tmpdir):
    """Test sampling with plots enabled"""
    output = str(tmpdir.mkdir("test"))
    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=100,
        plot=True,
        proposal_plots=True,
    )
    fs.run(plot=True)
    assert os.path.exists(os.path.join(output, "proposal", "pool_0.png"))
    assert os.path.exists(os.path.join(output, "proposal", "pool_0_log_q.png"))


@pytest.mark.slow_integration_test
def test_truncate_log_q(integration_model, tmpdir):
    """Test sampling with truncate_log_q"""
    output = str(tmpdir.mkdir("test"))
    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=500,
        plot=False,
        proposal_plots=False,
        constant_volume_mode=False,
        latent_prior="flow",
        truncate_log_q=True,
    )
    fs.run(plot=False)
    assert fs.ns.finalised


@pytest.mark.slow_integration_test
def test_prior_sampling(integration_model, tmpdir):
    """Test prior sampling"""
    output = str(tmpdir.mkdir("prior_sampling"))
    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=100,
        plot=False,
        prior_sampling=True,
    )
    fs.run(plot=False)

    assert len(fs.nested_samples) == 100
    assert np.isfinite(fs.logZ)


@pytest.mark.slow_integration_test
def test_sampling_resume_finalised(integration_model, tmp_path):
    """Test resuming the sampler after it finishes sampling"""
    output = tmp_path / "output"
    output.mkdir()
    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=100,
        plot=False,
    )
    fs.run(save=False, plot=False)
    assert os.path.exists(fs.ns.resume_file)

    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=100,
        plot=False,
    )

    fs.run(save=False, plot=False)


@pytest.mark.slow_integration_test
def test_debug_log_level(integration_model, tmpdir):
    """Test running with debug log-level."""
    logger = logging.getLogger("nessai")
    original_level = logger.level
    logger.setLevel("DEBUG")
    output = str(tmpdir.mkdir("debug_logging"))
    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=100,
        plot=False,
    )
    fs.run(plot=False)
    logger.setLevel(original_level)


@pytest.mark.slow_integration_test
def test_disable_vectorisation(tmp_path):
    """Assert vectorisation can be disabled"""

    class TestModel(IntegrationTestModel):
        def log_likelihood(self, x):
            # AssertionError won't be caught by nessai
            assert not (x.size > 1)
            return super().log_likelihood(x)

    output = tmp_path / "disable_vec"
    output.mkdir()

    model = TestModel()

    fs = FlowSampler(
        model,
        output=output,
        nlive=100,
        disable_vectorisation=True,
        plot=False,
    )
    fs.run(plot=False)


@pytest.mark.slow_integration_test
def test_likelihood_chunksize(tmp_path):
    """Assert likelihood chunksize limits number of samples in each call"""

    class TestModel(IntegrationTestModel):
        def log_likelihood(self, x):
            # AssertionError won't be caught by nessai
            assert not (x.size > self.likelihood_chunksize)
            return super().log_likelihood(x)

    output = tmp_path / "disable_vec"
    output.mkdir()

    model = TestModel()

    fs = FlowSampler(
        model, output=output, nlive=100, plot=False, likelihood_chunksize=10
    )
    fs.run(plot=False, save=False)


@pytest.mark.slow_integration_test
def test_allow_multi_valued_likelihood(tmp_path):
    """Assert a multi valued likelihood can be sampled from"""

    class TestModel(IntegrationTestModel):
        def log_likelihood(self, x):
            return super().log_likelihood(x) + 1e-10 * np.random.randn(x.size)

    output = tmp_path / "multi_value"
    output.mkdir()

    model = TestModel()

    fs = FlowSampler(
        model,
        output=output,
        nlive=100,
        plot=False,
        allow_multi_valued_likelihood=True,
    )
    fs.run(plot=False, save=False)


@pytest.mark.integration_test
def test_invalid_keyword_argument(integration_model, tmp_path):
    """Assert an error is raised if a keyword argument is unknown"""

    output = tmp_path / "kwargs_error"
    output.mkdir()

    with pytest.raises(
        RuntimeError,
        match="Unknown kwargs for FlowProposal: {'not_a_valid_kwarg'}.",
    ):
        FlowSampler(
            integration_model,
            output=output,
            nlive=100,
            plot=False,
            not_a_valid_kwarg=True,
        )


@pytest.mark.parametrize("extension", ["hdf5", "h5", "json"])
@pytest.mark.slow_integration_test
def test_sampling_result_extension(integration_model, tmp_path, extension):
    """Assert the correct extension is used"""
    output = tmp_path / "test"
    output.mkdir()
    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=100,
        plot=False,
        proposal_plots=False,
        result_extension=extension,
    )
    fs.run(plot=False)
    assert os.path.exists(os.path.join(output, f"result.{extension}"))


@pytest.mark.slow_integration_test
def test_sampling_with_checkpoint_callback(integration_model, tmp_path):
    """Test the usage if the checkpoint callbacks"""
    import pickle

    output = tmp_path / "test_callbacks"
    output.mkdir()

    checkpoint_file = output / "test.pkl"

    def callback(state):
        with open(checkpoint_file, "wb") as f:
            pickle.dump(state, f)

    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=100,
        plot=False,
        proposal_plots=False,
        checkpoint_callback=callback,
        max_iteration=100,
        checkpoint_on_iteration=True,
        checkpoint_interval=50,
    )
    fs.run(plot=False)
    assert fs.ns.iteration == 100

    del fs

    with open(checkpoint_file, "rb") as f:
        resume_data = pickle.load(f)

    fs = FlowSampler(
        integration_model,
        output=output,
        nlive=100,
        plot=False,
        proposal_plots=False,
        checkpoint_callback=callback,
        checkpoint_on_iteration=True,
        checkpoint_interval=50,
        resume_data=resume_data,
        resume=True,
    )
    fs.ns.max_iteration = 200
    fs.run()
    assert fs.ns.iteration == 200



# -*- coding: utf-8 -*-
"""
Global configuration for nessai.
"""
from dataclasses import dataclass, field
from typing import List, Tuple

import numpy as np


@dataclass
class LivepointsConfig:
    """Configuration for live points."""

    logl_dtype: str = "f8"
    """Default log-likelihood dtype"""
    it_dtype: str = "i4"
    """Default dtype for iteration parameter"""
    it_default: int = 0
    """Default value for the iteration parameter"""
    default_float_dtype: str = "f8"
    """Default dtype for parameters"""
    default_float_value: float = np.nan
    """Default value for parameters"""
    core_parameters: List[str] = field(
        default_factory=lambda: ["logP", "logL", "it"]
    )
    """List of the core non-sampling parameters included in all live points"""
    extra_parameters: List[str] = field(default_factory=lambda: [])
    """Additional extra parameters included in live points"""
    extra_parameters_dtype: List[str] = field(default_factory=lambda: [])
    """Defaults dtype for extra parameters"""
    extra_parameters_defaults: Tuple = field(default_factory=lambda: ())
    """Default values for additional extra extra"""

    _core_parameter_dtype: List[str] = None
    _core_parameter_defaults: Tuple = None
    _non_sampling_defaults: List = None
    _non_sampling_parameters: Tuple = None
    _non_sampling_dtype: List[str] = None

    @property
    def core_parameters_dtype(self) -> List[str]:
        """List of dtypes for the core non-sampling parameters"""
        if self._core_parameter_dtype is None:
            self._core_parameter_dtype = [
                self.default_float_dtype,
                self.logl_dtype,
                self.it_dtype,
            ]
        return self._core_parameter_dtype

    @property
    def core_parameters_defaults(self) -> Tuple:
        """Tuple of default values for core non-sampling parameters."""
        if self._core_parameter_defaults is None:
            self._core_parameter_defaults = (
                self.default_float_value,
                self.default_float_value,
                self.it_default,
            )
        return self._core_parameter_defaults

    @property
    def non_sampling_parameters(self) -> List[str]:
        """List of all the non-sampling parameters"""
        if self._non_sampling_parameters is None:
            self._non_sampling_parameters = (
                self.core_parameters + self.extra_parameters
            )
        return self._non_sampling_parameters

    @property
    def non_sampling_defaults(self) -> Tuple:
        """List of default values for all the non-sampling parameters"""
        if self._non_sampling_defaults is None:
            self._non_sampling_defaults = (
                self.core_parameters_defaults + self.extra_parameters_defaults
            )
        return self._non_sampling_defaults

    @property
    def non_sampling_dtype(self) -> List[str]:
        """List of the dtype for all the non-sampling parameters"""
        if self._non_sampling_dtype is None:
            self._non_sampling_dtype = (
                self.core_parameters_dtype + self.extra_parameters_dtype
            )
        return self._non_sampling_dtype

    def reset(self) -> None:
        """Reset the extra parameters and properties"""
        self.extra_parameters = []
        self.extra_parameters_defaults = ()
        self.extra_parameters_dtype = []
        self.reset_properties()

    def reset_properties(self) -> None:
        """Reset the cached properties"""
        self._core_parameter_dtype = None
        self._core_parameter_defaults = None
        self._non_sampling_defaults = None
        self._non_sampling_parameters = None
        self._non_sampling_dtype = None


@dataclass
class PlottingConfig:
    """Configuration for plotting."""

    disable_style: bool = False
    """Disable nessai's custom plotting style globally.

    Useful since all plotting functions use the
    :py:func:`~nessai.plot.nessai_style` decorator by default.
    """
    sns_style: str = "ticks"
    """Default seaborn style."""
    base_colour: str = "#02979d"
    """Base colour for plots."""
    highlight_colour: str = "#f5b754"
    """Highlight colour for plots."""
    line_colours: List[str] = field(
        default_factory=lambda: ["#4575b4", "#d73027", "#fad117", "#ff8c00"]
    )
    """Default line colours."""
    line_styles: List[str] = field(
        default_factory=lambda: ["-", "--", ":", "-."]
    )
    """Default line styles."""
    max_figsize: float = 50
    """Maximum figure size in either width or height.

    Based on the default DPI in matplotlib of 100, so this will give a maximum
    size of 5000 pixels.
    """


@dataclass
class GeneralConfig:
    """General configuration options"""

    eps: float = 1e-8
    """Epsilon value used for numerical stability"""


livepoints = LivepointsConfig()
plotting = PlottingConfig()
general = GeneralConfig()



# -*- coding: utf-8 -*-
"""
Test the properties in NestedSampler
"""
from collections import deque
import time
import datetime
import numpy as np
import pytest
from unittest.mock import MagicMock

from nessai.samplers.nestedsampler import NestedSampler


@pytest.fixture()
def sampler(sampler):
    sampler.state = MagicMock()
    return sampler


def test_log_evidence(sampler):
    """Check evidence is returned"""
    sampler.state.logZ = -2
    assert NestedSampler.log_evidence.__get__(sampler) == -2


def test_log_evidence_error(sampler):
    """Check the log-evidence error is returned"""
    sampler.state.log_evidence_error = 0.1
    assert NestedSampler.log_evidence_error.__get__(sampler) == 0.1


def test_information(sampler):
    """Check most recent information estimate is returned"""
    sampler.state.info = [1, 2, 3]
    assert NestedSampler.information.__get__(sampler) == 3


def test_population_time(sampler):
    """Assert the time is the some of the time for individual proposals"""
    sampler._uninformed_proposal = MagicMock()
    sampler._flow_proposal = MagicMock()
    sampler._uninformed_proposal.population_time = 1
    sampler._flow_proposal.population_time = 2
    assert NestedSampler.proposal_population_time.__get__(sampler) == 3


def test_acceptance(sampler):
    """Test the acceptance calculation"""
    sampler.iteration = 10
    sampler.likelihood_calls = 100
    assert NestedSampler.acceptance.__get__(sampler) == 0.1


def test_current_sampling_time(sampler):
    """Test the current sampling time"""
    sampler.finalised = False
    sampler.sampling_time = datetime.timedelta(seconds=10)
    sampler.sampling_start_time = datetime.datetime.now()
    time.sleep(0.01)
    t = NestedSampler.current_sampling_time.__get__(sampler)
    assert t.total_seconds() > 10.0


def test_current_sampling_time_finalised(sampler):
    """Test the current sampling time if the sampling has been finalised"""
    sampler.finalised = True
    sampler.sampling_time = 10
    assert NestedSampler.current_sampling_time.__get__(sampler) == 10


def test_last_updated(sampler):
    """Assert last training iteration is returned"""
    sampler.history = dict(training_iterations=[10, 20])
    assert NestedSampler.last_updated.__get__(sampler) == 20


def test_last_updated_no_training(sampler):
    """Assert None is return if the flow has not been trained"""
    sampler.history = dict(training_iterations=[])
    assert NestedSampler.last_updated.__get__(sampler) == 0


def test_last_updated_no_history(sampler):
    """Assert None is return if the flow has not been trained"""
    sampler.history = None
    assert NestedSampler.last_updated.__get__(sampler) == 0


def test_mean_acceptance(sampler):
    """Assert the mean is returned"""
    sampler.acceptance_history = [1.0, 2.0, 3.0]
    assert NestedSampler.mean_acceptance.__get__(sampler) == 2.0


def test_mean_acceptance_empty(sampler):
    """Assert nan is returned if no points have been proposed"""
    sampler.acceptance_history = deque(maxlen=10)
    assert np.isnan(NestedSampler.mean_acceptance.__get__(sampler))


def test_posterior_effective_sample_size(sampler):
    """Assert the state property is called"""
    sampler.state.effective_n_posterior_samples = 10
    out = NestedSampler.posterior_effective_sample_size.__get__(sampler)
    assert out == 10


def test_birth_log_likelihoods(sampler):
    """Assert the birth log-likelihoods are correct"""
    sampler.state.logLs = [-np.inf, 1, 2, 3, 4]
    dtype = [("it", "i4")]
    sampler.nested_samples = [
        np.array(
            [
                0,
            ],
            dtype=dtype,
        ),
        np.array(
            [
                1,
            ],
            dtype=dtype,
        ),
        np.array(
            [
                2,
            ],
            dtype=dtype,
        ),
        np.array(
            [
                0,
            ],
            dtype=dtype,
        ),
        np.array(
            [
                0,
            ],
            dtype=dtype,
        ),
        np.array(
            [
                4,
            ],
            dtype=dtype,
        ),
        np.array(
            [
                3,
            ],
            dtype=dtype,
        ),
    ]

    expected = np.array([-np.inf, 1, 2, -np.inf, -np.inf, 4, 3])

    out = NestedSampler.birth_log_likelihoods.__get__(sampler)

    np.testing.assert_array_equal(out, expected)



#!/usr/bin/env python

# Example of using nessai for a Rosenbrock likelihood.
# Also shows how to configure the normalising flow.

import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.utils import setup_logger

output = "./outdir/rosenbrock/"
logger = setup_logger(output=output)


class RosenbrockModel(Model):
    """Rosenbrock function defined in n dimensions on [-5, 5]^n.

    Based on the example in cpnest:\
        https://github.com/johnveitch/cpnest/blob/master/examples/rosenbrock.py
    """

    def __init__(self, dims):
        self.names = [f"x_{d}" for d in range(dims)]
        self.bounds = {n: [-5.0, 5.0] for n in self.names}

    def log_prior(self, x):
        """Uniform prior"""
        log_p = np.log(self.in_bounds(x), dtype="float")
        for bounds in self.bounds.values():
            log_p -= np.log(bounds[1] - bounds[0])
        return log_p

    def log_likelihood(self, x):
        """Log-likelihood"""
        # Unstructured view returns a view of the inputs as a "normal" numpy
        # array without any fields. It will included the parameters listed in
        # names and in the same order. This is faster then converting from
        # live points to an array.
        x = self.unstructured_view(x)
        return -np.sum(
            100.0 * (x[..., 1:] - x[..., :-1] ** 2.0) ** 2.0
            + (1.0 - x[..., :-1]) ** 2.0,
            axis=-1,
        )


model = RosenbrockModel(5)

# The Rosenbrock likelihood is more complex, so we configure the normalising
# flow to improve nessai's performance.
flow_config = dict(model_config=dict(n_blocks=4, n_neurons=10, n_layers=3))

# Configure the sampler.
fs = FlowSampler(
    model,
    output=output,
    flow_config=flow_config,
    resume=False,
    seed=1451,
)

# And go!
fs.run()



# -*- coding: utf-8 -*-
"""
General utilities for nessai.
"""
from .distributions import get_multivariate_normal, get_uniform_distribution
from .hist import auto_bins
from .indices import bonferroni_correction, compute_indices_ks_test
from .io import (
    NessaiJSONEncoder,
    is_jsonable,
    safe_file_dump,
    save_dict_to_hdf5,
    save_live_points,
    save_to_json,
)
from .logging import setup_logger
from .rescaling import (
    configure_edge_detection,
    detect_edge,
    determine_rescaled_bounds,
    inverse_rescale_minus_one_to_one,
    inverse_rescale_zero_to_one,
    logit,
    rescale_minus_one_to_one,
    rescale_zero_to_one,
    rescaling_functions,
    sigmoid,
)
from .sampling import (
    compute_radius,
    draw_gaussian,
    draw_nsphere,
    draw_surface_nsphere,
    draw_truncated_gaussian,
    draw_uniform,
)
from .distance import compute_minimum_distances
from .stats import rolling_mean
from .structures import replace_in_list
from .threading import configure_threads


__all__ = [
    "BoxUniform",
    "MultivariateNormal",
    "NessaiJSONEncoder",
    "auto_bins",
    "bonferroni_correction",
    "compute_indices_ks_test",
    "compute_minimum_distances",
    "compute_radius",
    "configure_edge_detection",
    "configure_threads",
    "detect_edge",
    "determine_rescaled_bounds",
    "distance",
    "distributions",
    "draw_gaussian",
    "draw_nsphere",
    "draw_surface_nsphere",
    "draw_truncated_gaussian",
    "draw_uniform",
    "get_multivariate_normal",
    "get_uniform_distribution",
    "hist",
    "indices",
    "inverse_rescale_minus_one_to_one",
    "inverse_rescale_zero_to_one",
    "is_jsonable",
    "logit",
    "rolling_mean",
    "replace_in_list",
    "rescale_minus_one_to_one",
    "rescale_zero_to_one",
    "rescaling",
    "rescaling_functions",
    "safe_file_dump",
    "sampling",
    "save_dict_to_hdf5",
    "save_live_points",
    "save_to_json",
    "setup_logger",
    "sigmoid",
    "spatial",
    "structures",
    "threading",
]



# -*- coding: utf-8 -*-
"""nessai: Nested sampling with Artificial Intelligence

nessai is a nested sampling algorithm for Bayesian Inference that incorporates
normalising flows. It is designed for applications where the Bayesian
likelihood is computationally expensive.
"""
import logging
from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    # package is not installed
    __version__ = "unknown"

logging.getLogger(__name__).addHandler(logging.NullHandler())



# -*- coding: utf-8 -*-
"""
Combined reparameterisation.
"""
import logging

import numpy as np

from ..utils.sorting import sort_reparameterisations

logger = logging.getLogger(__name__)


class CombinedReparameterisation(dict):
    """Class to handle multiple reparameterisations

    Parameters
    ----------
    reparameterisations : list, optional
        List of reparameterisations to add to the combined reparameterisations.
        Further reparameterisations can be added using
        `add_reparameterisations`.
    reverse_order : bool
        If True the order of the reparameterisations will be reversed compared
        to the default ordering.
    """

    def __init__(self, reparameterisations=None, reverse_order=False):
        super().__init__()
        self.reparameterisations = {}
        self.parameters = []
        self.prime_parameters = []
        self.requires = []
        self.order = []
        self.reverse_order = reverse_order
        if reparameterisations is not None:
            self.add_reparameterisations(reparameterisations)

    @property
    def has_prime_prior(self):
        """Boolean to check if prime prior can be enabled"""
        return all(r.has_prime_prior for r in self.values())

    @property
    def requires_prime_prior(self):
        """Boolean to check if any of the priors require the prime space"""
        return any(r.requires_prime_prior for r in self.values())

    @property
    def to_prime_order(self):
        """Order when converting to the prime space"""
        if self.reverse_order:
            return reversed(self.order)
        else:
            return self.order

    @property
    def from_prime_order(self):
        """Order when converting from the prime space"""
        if self.reverse_order:
            return self.order
        else:
            return reversed(self.order)

    def _add_reparameterisation(self, reparameterisation):
        requires = reparameterisation.requires
        if requires and (
            any([req not in self.parameters for req in requires])
            and any([req not in self.prime_parameters for req in requires])
        ):
            raise RuntimeError(
                f"Could not add {reparameterisation}, missing requirement(s): "
                f"{reparameterisation.requires}. Current: {self.parameters}."
            )

        self[reparameterisation.name] = reparameterisation
        self.order = list(self.keys())
        self.parameters += reparameterisation.parameters
        self.prime_parameters += reparameterisation.prime_parameters
        self.requires += reparameterisation.requires

    def add_reparameterisation(self, reparameterisation):
        """Add a reparameterisation"""
        self.add_reparameterisations(reparameterisation)

    def add_reparameterisations(self, reparameterisations):
        """Add multiple reparameterisations

        Parameters
        ----------
        reparameterisations : list of :`obj`:Reparameterisation
            List of reparameterisations to add.
        """
        if not isinstance(reparameterisations, list):
            reparameterisations = [reparameterisations]

        logger.debug("Sorting reparameterisations")
        logger.debug(f"Existing parameters: {self.parameters}")
        reparameterisations = sort_reparameterisations(
            reparameterisations,
            existing_parameters=self.parameters,
        )

        for r in reparameterisations:
            self._add_reparameterisation(r)

    def check_order(self):
        """Check the order of the reparameterisations is valid.

        Raises
        ------
        RuntimeError
            Raised if the order is invalid.
        """
        parameters = []
        for key in self.from_prime_order:
            if not all([r in parameters for r in self[key].requires]):
                raise RuntimeError(
                    "Order of reparameterisations is invalid (x' -> x)"
                    f"{self[key].name} requires {self[key].requires} but with "
                    f"the current order only the parameters {parameters} "
                    "would be available."
                )
            parameters += self[key].parameters

    def reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated
        """
        for key in self.to_prime_order:
            x, x_prime, log_j = self[key].reparameterise(
                x, x_prime, log_j, **kwargs
            )
        return x, x_prime, log_j

    def inverse_reparameterise(self, x, x_prime, log_j, **kwargs):
        """
        Apply the reparameterisation to convert from x-space
        to x'-space

        Parameters
        ----------
        x : structured array
            Array
        x_prime : structured array
            Array to be update
        log_j : array_like
            Log jacobian to be updated
        """
        for key in self.from_prime_order:
            x, x_prime, log_j = self[key].inverse_reparameterise(
                x, x_prime, log_j, **kwargs
            )
        return x, x_prime, log_j

    def update_bounds(self, x):
        """
        Update the bounds used for the reparameterisation
        """
        for r in self.values():
            if hasattr(r, "update_bounds"):
                logger.debug(f"Updating bounds for: {r.name}")
                r.update_bounds(x)

    def reset_inversion(self):
        """
        Reset edges for boundary inversion
        """
        for r in self.values():
            if hasattr(r, "reset_inversion"):
                r.reset_inversion()

    def update(self, x):
        """Update the reparameterisations given a set of points."""
        for r in self.values():
            r.update(x)

    def log_prior(self, x):
        """
        Compute any additional priors for auxiliary parameters
        """
        log_p = np.zeros(x.size)
        for r in self.values():
            if r.has_prior:
                log_p += r.log_prior(x)
        return log_p

    def x_prime_log_prior(self, x_prime):
        """
        Compute the prior in the prime space
        """
        log_p = np.zeros(x_prime.size)
        for r in self.values():
            log_p += r.x_prime_log_prior(x_prime)
        return log_p



"""Test training functions in ImportanceFlowProposal"""

import os

from nessai.proposal.importance import ImportanceFlowProposal as IFP
from nessai.livepoint import numpy_array_to_live_points
from nessai.utils.testing import assert_structured_arrays_equal
import numpy as np
from unittest.mock import MagicMock, Mock, patch
import pytest


@pytest.fixture
def ifp(ifp, tmp_path):
    ifp.output = tmp_path / "test_train"
    ifp.level_count = 2
    ifp.weighted_kl = False
    ifp._reset_flow = False
    ifp._weights = {-1: 0.25, 0: 0.25, 1: 0.25, 2: 0.25}
    return ifp


@pytest.fixture(params=[True, False])
def reset(request):
    return request.param


def test_train_basic(ifp, x, x_prime, reset):
    """Test the main training components"""
    ifp.rescale = Mock(return_value=(x_prime, None))
    ifp._reset_flow = reset
    ifp.plot_training = True

    ifp.flow = Mock()
    ifp.flow.models = [object(), object(), object()]

    def add_flow(reset=False):
        ifp.flow.models.append(object())

    ifp.flow.add_new_flow = Mock(side_effect=add_flow)

    IFP.train(ifp, samples=x, plot=False, test=True)

    # Number of levels should be level index +1
    assert len(ifp.flow.models) == (ifp.level_count + 1)

    ifp.flow.train.assert_called_once_with(
        x_prime,
        weights=None,
        output=os.path.join(ifp.output, "level_3", ""),
        plot=True,
        test=True,  # Make sure kwargs are passed
    )

    ifp.flow.add_new_flow.assert_called_once_with(reset=reset)
    assert len(ifp._weights) == (ifp.level_count + 2)
    assert ifp._weights[ifp.level_count] is np.nan


@patch("nessai.proposal.importance.plot_live_points")
@patch("nessai.proposal.importance.plot_1d_comparison")
@patch("nessai.proposal.importance.plot_histogram")
@pytest.mark.usefixtures("ins_parameters")
def test_train_plotting(
    mock_hist,
    mock_plot_1d,
    mock_plot_live_points,
    ifp,
    x,
    x_prime,
    model,
):
    """Test plotting in training function"""

    names = model.names
    level_output = os.path.join(ifp.output, "level_3", "")

    n = 10
    weights = np.random.rand(n)
    prime_samples = np.random.randn(n, len(names))
    samples = numpy_array_to_live_points(prime_samples, names)
    log_prob = np.random.rand(n)
    log_j = np.random.randn(n)

    samples["logQ"] = log_prob

    ifp.rescale = Mock(return_value=(x_prime, None))
    ifp.inverse_rescale = Mock(return_value=(samples, log_j))
    ifp.flow = MagicMock()
    ifp.flow.sample_and_log_prob = Mock(return_value=(prime_samples, log_prob))

    IFP.train(ifp, samples=x, plot=True, weights=weights)

    mock_hist.assert_called_once()
    np.testing.assert_array_equal(
        mock_hist.call_args[0][0], weights / np.sum(weights)
    )
    assert mock_hist.call_args_list[0][1]["filename"] == os.path.join(
        level_output, "training_weights.png"
    )

    mock_plot_1d.assert_called_once_with(
        x_prime,
        convert_to_live_points=True,
        filename=os.path.join(level_output, "prime_training_data.png"),
    )

    assert_structured_arrays_equal(
        mock_plot_live_points.call_args_list[0][0],
        x,
    )

    assert_structured_arrays_equal(
        mock_plot_live_points.call_args_list[1][0],
        samples,
    )

    print(mock_plot_live_points.call_args_list[0])
    assert mock_plot_live_points.call_args_list[0][1][
        "filename"
    ] == os.path.join(level_output, "training_data.png")
    assert mock_plot_live_points.call_args_list[1][1][
        "filename"
    ] == os.path.join(level_output, "generated_samples.png")


@pytest.mark.parametrize(
    "weights, weighted_kl", [[np.random.rand(10), False], [None, True]]
)
@pytest.mark.usefixtures("ins_parameters")
def test_train_weights(ifp, x, x_prime, weights, weighted_kl):

    x["logW"] = np.log(np.random.rand(x.size))

    ifp.rescale = Mock(return_value=(x_prime, None))
    ifp.flow = MagicMock()
    ifp.weighted_kl = weighted_kl
    ifp.plot_training = False

    if weights is not None:
        expected_weights = weights / np.sum(weights)
    else:
        expected_weights = np.exp(x["logW"]) / np.sum(np.exp(x["logW"]))

    IFP.train(ifp, samples=x, weights=weights, plot=False)

    actual_weights = ifp.flow.train.call_args[1]["weights"]
    np.testing.assert_almost_equal(np.sum(actual_weights), 1.0, decimal=10)
    np.testing.assert_array_almost_equal(actual_weights, expected_weights)


def test_train_weights_nan(ifp, x, x_prime):
    ifp.rescale = Mock(return_value=(x_prime, None))
    ifp.flow = MagicMock()
    ifp.weighted_kl = False
    ifp.plot_training = False
    weights = np.ones(x.size)
    weights[0] = np.nan
    with pytest.raises(ValueError, match=r"Weights contain NaN\(s\)"):
        IFP.train(ifp, samples=x, weights=weights, plot=False)


@pytest.mark.integration_test
@pytest.mark.usefixtures("ins_parameters")
def test_training_and_prob(model, tmp_path):
    ifp = IFP(
        model,
        output=tmp_path / "test_training_and_prob",
        weighted_kl=False,
    )
    ifp.initialise()
    weights = {-1: 1.0}
    for i in range(4):
        ifp.train(model.new_point(10), max_epochs=2)
        weights = {j - 1: 1 / (i + 2) for j in range(i + 2)}
        ifp.update_proposal_weights(weights)
        x, _ = ifp.draw(10)
    log_Q, log_q = ifp.compute_meta_proposal_samples(x)

    assert len(log_Q) == 10
    assert log_q.shape == (10, 5)



#!/usr/bin/env python

"""
Example of using `corner_plot` to plot the posterior distribution with truth
values.

Requires corner.
"""

import os
import numpy as np

from nessai.flowsampler import FlowSampler
from nessai.model import Model
from nessai.plot import corner_plot
from nessai.utils import setup_logger


output = "./outdir/corner_plot_example/"
logger = setup_logger(output=output)


# Generate the data
truth = {"mu": 1.7, "sigma": 0.7}
bounds = {"mu": [-3, 3], "sigma": [0.01, 3]}
n_points = 1000
data = np.random.normal(truth["mu"], truth["sigma"], size=n_points)


class GaussianLikelihood(Model):
    """
    Gaussian likelihood with the mean and standard deviation as the parameters
    to infer.

    Parameters
    ----------
    data : :obj:`numpy.ndarray`
        Array of data.
    bounds : dict
        The prior bounds.
    """

    def __init__(self, data, bounds):
        self.names = list(bounds.keys())
        self.bounds = bounds
        self.data = data

    def log_prior(self, x):
        """Uniform prior on both parameters."""
        log_p = np.log(self.in_bounds(x), dtype="float")
        for bounds in self.bounds.values():
            log_p -= np.log(bounds[1] - bounds[0])
        return log_p

    def log_likelihood(self, x):
        """Gaussian likelihood."""
        log_l = np.sum(
            -np.log(x["sigma"])
            - 0.5 * ((self.data - x["mu"]) / x["sigma"]) ** 2
        )
        return log_l


fs = FlowSampler(
    GaussianLikelihood(data, bounds),
    output=output,
    resume=False,
    seed=1234,
)

fs.run(plot=False)

# Produce a corner plot that includes the true values for each parameter and
# uses custom labels for each.
corner_plot(
    fs.posterior_samples,
    include=list(truth.keys()),
    truths=list(truth.values()),
    labels=[r"$\mu$", r"$\sigma$"],
    filename=os.path.join(output, "posterior_w_truth.png"),
)



from unittest.mock import create_autospec

import pytest
from nessai.evidence import _INSIntegralState
from nessai.proposal.importance import ImportanceFlowProposal
from nessai.samplers.importancesampler import (
    ImportanceNestedSampler,
    OrderedSamples,
)
from nessai.livepoint import (
    add_extra_parameters_to_live_points,
    reset_extra_live_points_parameters,
)
from nessai.model import Model
import numpy as np
from scipy.special import logsumexp


@pytest.fixture(scope="module", autouse=True)
def ins_livepoint_params():
    reset_extra_live_points_parameters()
    add_extra_parameters_to_live_points(["logW", "logQ"])
    # Test happens here
    yield

    # Called after every test
    reset_extra_live_points_parameters()


@pytest.fixture(scope="module", params=[False, True])
def iid(request):
    return request.param


@pytest.fixture
def ins():
    obj = create_autospec(ImportanceNestedSampler)
    obj.model = create_autospec(Model)
    obj.training_samples = create_autospec(OrderedSamples)
    obj.training_samples.state = create_autospec(_INSIntegralState)
    obj.iid_samples = None
    return obj


@pytest.fixture()
def proposal():
    return create_autospec(ImportanceFlowProposal)


@pytest.fixture(scope="module")
def n_it():
    return 10


@pytest.fixture(scope="module")
def n_samples():
    return 1000


@pytest.fixture
def samples(model, n_samples, n_it, log_q):
    x = model.sample_unit_hypercube(n_samples)
    x["it"] = np.random.randint(-1, n_it - 1, size=len(x))
    x["logL"] = model.log_likelihood(x)
    x["logP"] = model.log_prior(x)

    alpha = np.bincount(
        x["it"] + np.abs(x["it"].min()), minlength=n_it
    ).astype(float)
    alpha /= alpha.sum()
    x["logQ"] = logsumexp(log_q, axis=1, b=alpha)
    x["logW"] = -x["logQ"].copy()
    return x


@pytest.fixture
def log_q(n_samples, n_it):
    return np.random.randn(n_samples, n_it)


@pytest.fixture
def history(n_it):
    keys = [
        "logX",
        "logZ",
        "min_log_likelihood",
        "max_log_likelihood",
        "median_logL",
        "likelihood_evaluations",
        "n_post",
        "logL_threshold",
        "n_added",
        "n_removed",
        "n_live",
        "n_post",
        "live_points_ess",
        "leakage_live_points",
        "leakage_new_points",
        "gradients",
        "samples_entropy",
        "proposal_entropy",
    ]
    d = {k: np.random.randn(n_it).tolist() for k in keys}
    d["stopping_criteria"] = dict(
        log_dZ=np.arange(n_it),
        ratio=np.arange(n_it),
        ess=np.arange(n_it),
        kl=np.arange(n_it),
    )
    return d



# -*- coding: utf-8 -*-
"""
Test io utilities.
"""
import os
import json
import h5py
import numpy as np
import pickle
import pytest
from unittest.mock import call, create_autospec, mock_open, patch

from nessai import config
from nessai.livepoint import numpy_array_to_live_points
from nessai.utils.io import (
    NessaiJSONEncoder,
    add_dict_to_hdf5_file,
    encode_for_hdf5,
    is_jsonable,
    safe_file_dump,
    save_dict_to_hdf5,
    save_live_points,
    save_to_json,
)


@pytest.fixture
def data_dict():
    data = dict(
        a=np.array([1, 2, 3]),
        b=np.array([(1, 2)], dtype=[("x", "f4"), ("y", "f4")]),
        cls=object(),
        l=[1, 2, 3],
        dict1={"a": None, "b": 2},
        dict2={"c": [1, 2, 3], "array": np.array([3, 4, 5])},
        s="A string",
        nan=None,
    )
    return data


def test_is_jsonable_true():
    """Assert True is return is json.dumps does not raise an error"""
    assert is_jsonable({"x": 2})


@pytest.mark.parametrize("cls", [TypeError, OverflowError])
def test_is_jsonable_false(
    cls,
):
    """Assert True is return is json.dumps does not raise an error"""
    with patch("json.dumps", side_effect=cls()) as mock:
        assert not is_jsonable({"x": 2})
    mock.assert_called_once_with({"x": 2})


@pytest.mark.parametrize(
    "input, expected",
    [(np.int32(2), 2), (np.float64(2), 2.0), (np.array([1, 2]), [1, 2])],
)
def test_JSON_encoder_numpy(input, expected):
    """Test the JSON encoder with numpy inputs"""
    e = create_autospec(NessaiJSONEncoder)
    output = NessaiJSONEncoder.default(e, input)
    assert output == expected


def test_JSON_encoder_object():
    """Test the JSON with an object that is JSONable"""
    e = create_autospec(NessaiJSONEncoder)
    input = object()
    with patch("nessai.utils.io.is_jsonable", return_value=False) as m:
        output = NessaiJSONEncoder.default(e, input)
    m.assert_called_once_with(input)
    assert "object object at" in output


def test_JSON_encoder_other():
    """Test the JSON with an object that is JSONable"""
    e = create_autospec(NessaiJSONEncoder)
    input = "Hello"
    with patch("json.JSONEncoder.default", return_value="Hello") as m:
        output = NessaiJSONEncoder.default(e, input)
    m.assert_called_once_with(input)
    assert output == "Hello"


def test_save_to_json():
    """Assert json.dump is called with the correct arguments"""
    d = dict(a=1)
    expected_kwargs = dict(indent=4, cls=NessaiJSONEncoder, test=True)
    mop = mock_open()
    filename = "test.json"
    with patch("builtins.open", mop, create=True), patch(
        "json.dump"
    ) as mock_dump:
        save_to_json(d, filename, test=True)
    fp = mop()
    mock_dump.assert_called_once_with(d, fp, **expected_kwargs)


@pytest.mark.integration_test
def test_save_to_json_integration(tmp_path, data_dict):
    """Integration test for save to json"""
    filename = tmp_path / "result.json"
    save_to_json(data_dict, filename)
    assert os.path.exists(filename)

    with open(filename, "r") as fp:
        out = json.load(fp)
    assert list(data_dict.keys()) == list(out.keys())


def test_safe_file_dump():
    """Test safe file dump."""
    m = mock_open()
    data = np.array([1, 2])
    with patch("builtins.open", m) as mo, patch("pickle.dump") as md, patch(
        "shutil.move"
    ) as msm:
        safe_file_dump(data, "test.pkl", pickle, save_existing=False)
    mo.assert_called_once_with("test.pkl.temp", "wb")
    np.testing.assert_array_equal(md.call_args_list[0][0][0], data)
    msm.assert_called_once_with("test.pkl.temp", "test.pkl")


def test_safe_file_dump_save_existing():
    """Test safe file dump."""
    import pickle

    m = mock_open()
    data = np.array([1, 2])
    with patch("os.path.exists", return_value=True) as mpe, patch(
        "builtins.open", m
    ) as mo, patch("pickle.dump") as md, patch("shutil.move") as msm:
        safe_file_dump(data, "test.pkl", pickle, save_existing=True)
    mpe.assert_called_once_with("test.pkl")
    mo.assert_called_once_with("test.pkl.temp", "wb")
    np.testing.assert_array_equal(md.call_args_list[0][0][0], data)
    msm.assert_has_calls(
        [
            call("test.pkl", "test.pkl.old"),
            call("test.pkl.temp", "test.pkl"),
        ]
    )


def test_safe_file_dump_integration(tmp_path):
    """Integration test for safe file dump"""
    path = tmp_path
    f = path / "test.pkl"
    f.write_text("a")
    data = "b"
    path = str(path)
    safe_file_dump(
        data, os.path.join(path, "test.pkl"), pickle, save_existing=True
    )
    assert os.path.exists(os.path.join(path, "test.pkl"))
    assert os.path.exists(os.path.join(path, "test.pkl.old"))


def test_save_live_points(tmp_path):
    """Test the function for saving live points"""
    d = {"x": [1, 2], "y": [3, 4]}
    d.update(
        {
            k: 2 * [v]
            for k, v in zip(
                config.livepoints.non_sampling_parameters,
                config.livepoints.non_sampling_defaults,
            )
        }
    )
    live_points = numpy_array_to_live_points(
        np.array([[1, 3], [2, 4]]), ["x", "y"]
    )
    filename = os.path.join(str(tmp_path), "test.json")
    save_live_points(live_points, filename)

    with open(filename, "r") as fp:
        d_out = json.load(fp)

    np.testing.assert_equal(d_out, d)


@pytest.mark.parametrize(
    "value, expected", [(None, "__none__"), ([1, 2], [1, 2])]
)
def test_encode_to_hdf5(value, expected):
    """Assert values are correctly encoded."""
    assert encode_for_hdf5(value) == expected


def test_add_dict_to_hdf5_file(tmp_path, data_dict):
    """Assert dictionaries is correctly converted"""
    data_dict.pop("cls")
    with h5py.File(tmp_path / "test.h5", "w") as f:
        add_dict_to_hdf5_file(f, "/", data_dict)
        assert list(f.keys()) == sorted(data_dict.keys())
        assert f["/dict1/a"][()].decode() == "__none__"
        np.testing.assert_array_equal(
            f["dict2/array"][:], data_dict["dict2"]["array"]
        )


def test_save_dict_to_hdf5(data_dict):
    """Assert the correct arguments are specified"""
    f = mock_open()
    filename = "result.h5"
    with patch("h5py.File", f) as mock_file, patch(
        "nessai.utils.io.add_dict_to_hdf5_file"
    ) as mock_add:
        save_dict_to_hdf5(data_dict, filename)
    mock_file.assert_called_once_with(filename, "w")
    mock_add.assert_called_once_with(f(), "/", data_dict)


@pytest.mark.integration_test
def test_save_dict_to_hdf5_integration(tmp_path, data_dict):
    """Test saving a dict to HDF5 integration"""
    # Do not need to support saving a class to HDF5
    data_dict.pop("cls")
    filename = tmp_path / "result.hdf5"
    save_dict_to_hdf5(data_dict, filename)

    with h5py.File(filename, "r") as f:
        keys = list(f.keys())
    assert keys == sorted(list(data_dict.keys()))



# -*- coding: utf-8 -*-
"""
Test the distance converter classes.
"""
import numpy as np
import pytest
from scipy import stats
from unittest.mock import MagicMock, Mock, call, create_autospec, patch

from nessai.gw.utils import (
    ComovingDistanceConverter,
    DistanceConverter,
    NullDistanceConverter,
    PowerLawConverter,
    get_distance_converter,
)


@pytest.fixture
def base_converter():
    return create_autospec(DistanceConverter)


@pytest.fixture
def null_converter():
    return create_autospec(NullDistanceConverter)


@pytest.fixture
def power_law_converter():
    return create_autospec(PowerLawConverter)


@pytest.fixture
def comoving_vol_converter():
    return create_autospec(ComovingDistanceConverter)


def test_has_conversion():
    """Assert the values for has_conversion are correct"""
    assert DistanceConverter.has_conversion is False
    assert NullDistanceConverter.has_conversion is False
    assert PowerLawConverter.has_conversion is True
    assert ComovingDistanceConverter.has_conversion is True


def test_has_jacobian():
    """Assert the values of has_jacobian are corrects"""
    assert DistanceConverter.has_jacobian is False
    assert NullDistanceConverter.has_jacobian is True
    assert PowerLawConverter.has_jacobian is True
    assert ComovingDistanceConverter.has_jacobian is False


def test_converter_error():
    """Assert an error is raised if the methods are not implemented"""
    with pytest.raises(TypeError) as excinfo:
        DistanceConverter()
    assert "abstract methods" in str(excinfo.value)


def test_converter_to_uniform_parameter_error(base_converter):
    with pytest.raises(NotImplementedError):
        DistanceConverter.to_uniform_parameter(base_converter, None)


def test_converter_from_uniform_parameter_error(base_converter):
    with pytest.raises(NotImplementedError):
        DistanceConverter.from_uniform_parameter(base_converter, None)


def test_null_converter_init(null_converter, caplog):
    """Test the null distance converter init"""
    NullDistanceConverter.__init__(null_converter, d_min=10)
    assert "Kwargs {'d_min': 10} will be ignored" in caplog.text


def test_null_converter_to_uniform(null_converter):
    """Test the null distance converter to uniform"""
    d = np.arange(10)
    d_out, log_j = NullDistanceConverter.to_uniform_parameter(
        null_converter, d
    )
    np.testing.assert_array_equal(d_out, d)
    np.testing.assert_equal(log_j, 0)


def test_null_converter_from_uniform(null_converter):
    """Test the null distance converter from uniform"""
    d = np.arange(10)
    d_out, log_j = NullDistanceConverter.from_uniform_parameter(
        null_converter, d
    )
    np.testing.assert_array_equal(d_out, d)
    np.testing.assert_equal(log_j, 0)


def test_power_law_converter_missing_power():
    """Assert an error is raised if the power is not specified."""

    with pytest.raises(RuntimeError) as excinfo:
        PowerLawConverter(power=None)
    assert "Must specify the power" in str(excinfo.value)


@pytest.mark.parametrize("power, func", [(1, np.sqrt), (2, np.cbrt)])
def test_power_law_converter_init(power_law_converter, power, func):
    """Test the init method with powers of 1 or 2"""
    PowerLawConverter.__init__(power_law_converter, power=power, scale=200)
    assert power_law_converter._power == (power + 1)
    assert power_law_converter._f is func
    assert power_law_converter.scale == 200


@pytest.mark.parametrize("power", [3, 4, 5])
def test_power_law_converter_init_power_greater_than_2(
    power_law_converter, power
):
    """Test the init method for powers > 2"""
    x = 3
    PowerLawConverter.__init__(power_law_converter, power=power, scale=200)
    assert power_law_converter._power == (power + 1)
    assert power_law_converter._f(x) == (x ** (1 / (power + 1)))


@pytest.mark.parametrize(
    "d, scale, power, expected",
    [(1.0, 1.0, 3, np.log(4)), (1000.0, 1000.0, 2, -5.80914299)],
)
def test_power_law_jacobian(power_law_converter, d, scale, power, expected):
    """Test the power law Jacobian"""
    power_law_converter._power = power + 1
    power_law_converter.scale = scale
    log_j = PowerLawConverter._log_jacobian(power_law_converter, d)
    np.testing.assert_almost_equal(expected, log_j)


@pytest.mark.parametrize(
    "d, scale, power, expected",
    [(1.0, 1.0, 3, -np.log(4)), (1.0, 1000.0, 2, 5.80914299)],
)
def test_power_law_jacobian_inverse(
    power_law_converter, d, scale, power, expected
):
    """Test the power law Jacobian"""
    power_law_converter._power = power + 1
    power_law_converter.scale = scale
    log_j = PowerLawConverter._log_jacobian_inv(power_law_converter, d)
    np.testing.assert_almost_equal(expected, log_j)


def test_power_law_to_uniform_parameter(power_law_converter):
    """Test the conversion to a uniform parameter"""
    d = 20
    power_law_converter.scale = 10
    power_law_converter._power = 2
    power_law_converter._log_jacobian = Mock(return_value=2)
    du, lj = PowerLawConverter.to_uniform_parameter(power_law_converter, d)
    power_law_converter._log_jacobian.assert_called_once_with(d)
    assert du == 4
    assert lj == 2


def test_power_law_from_uniform_parameter(power_law_converter):
    """Test the conversion from a uniform parameter"""
    d = 4
    power_law_converter.scale = 10
    power_law_converter._f = Mock(return_value=2)
    power_law_converter._log_jacobian_inv = Mock(return_value=-2)
    du, lj = PowerLawConverter.from_uniform_parameter(power_law_converter, d)
    power_law_converter._f.assert_called_once_with(d)
    power_law_converter._log_jacobian_inv.assert_called_once_with(d)
    assert du == 20
    assert lj == -2


@pytest.mark.requires("astropy")
def test_comoving_vol_init(comoving_vol_converter):
    """Test the init method"""
    d_min = 100
    d_max = 1000
    dc_min = Mock()
    dc_min.value = 2
    dc_max = Mock()
    dc_max.value = 2000

    n_interp = 10
    dl_array = Mock()
    dl_array.value = np.arange(n_interp)
    dc_array = np.arange(n_interp)
    z_values = np.arange(n_interp + 2)

    # Patch hell - There has to be a better way
    with patch(
        "nessai.gw.utils.cosmo.Planck15", return_value=MagicMock()
    ), patch(
        "nessai.gw.utils.cosmo.Planck15.comoving_distance",
        side_effect=[dc_min, dc_max],
    ) as cd_mock, patch(
        "nessai.gw.utils.cosmo.Planck15.luminosity_distance",
        side_effect=[dl_array],
    ) as ld_mock, patch(
        "nessai.gw.utils.cosmo.z_at_value", side_effect=z_values
    ), patch(
        "numpy.linspace", return_value=dc_array
    ) as linspace_mock, patch(
        "nessai.gw.utils.interpolate.splrep", side_effect=["dc2dl", "dl2dc"]
    ) as interp_mock:
        ComovingDistanceConverter.__init__(
            comoving_vol_converter,
            d_min=d_min,
            d_max=d_max,
            n_interp=n_interp,
        )
    cd_mock.assert_has_calls([call(0), call(1)])
    ld_mock.assert_has_calls([call(z_values[2:].tolist())])

    assert comoving_vol_converter.dl_min == 95
    assert comoving_vol_converter.dl_max == 1050
    assert comoving_vol_converter.dc_min == 2
    assert comoving_vol_converter.dc_max == 2000
    linspace_mock.assert_called_once_with(2, 2000, 10)

    interp_mock.assert_has_calls(
        [call(dc_array, dl_array.value), call(dl_array.value, dc_array)]
    )
    assert comoving_vol_converter.interp_dc2dl == "dc2dl"
    assert comoving_vol_converter.interp_dl2dc == "dl2dc"


@pytest.mark.requires("astropy")
def test_comoving_vol_init_invalid_cosmology(comoving_vol_converter):
    """Test the init method when an invalid cosmology is given"""
    with pytest.raises(RuntimeError) as excinfo:
        ComovingDistanceConverter.__init__(
            comoving_vol_converter, cosmology="Planck"
        )
    assert "Could not get specified cosmology" in str(excinfo.value)


def test_comoving_vol_to_uniform(comoving_vol_converter):
    """Test converting to a uniform parameter"""
    d = 1000
    grid = Mock()
    comoving_vol_converter.interp_dl2dc = grid
    comoving_vol_converter.scale = 100
    with patch(
        "nessai.gw.utils.interpolate.splev", return_value=200
    ) as interp_mock:
        du, lj = ComovingDistanceConverter.to_uniform_parameter(
            comoving_vol_converter, d
        )
    interp_mock.assert_called_once_with(d, grid, ext=3)
    assert du == 8.0
    assert lj == 0


def test_comoving_vol_from_uniform(comoving_vol_converter):
    """Test converting from a uniform parameter"""
    d = 1000
    grid = Mock()
    comoving_vol_converter.interp_dc2dl = grid
    comoving_vol_converter.scale = 5
    with patch(
        "nessai.gw.utils.interpolate.splev", return_value=20
    ) as interp_mock:
        du, lj = ComovingDistanceConverter.from_uniform_parameter(
            comoving_vol_converter, d
        )
    # 5 * cbrt(1000) = 50
    interp_mock.assert_called_once_with(50, grid, ext=3)
    assert du == 20
    assert lj == 0


@pytest.mark.parametrize(
    "prior, cls",
    [
        ("uniform-comoving-volume", ComovingDistanceConverter),
        ("power-law", PowerLawConverter),
        ("other", NullDistanceConverter),
        (None, NullDistanceConverter),
    ],
)
def test_get_distance_converter(prior, cls):
    """Assert the correct class is returned."""
    cls_out = get_distance_converter(prior)
    assert cls_out == cls


@pytest.mark.parametrize("power", [1, 2, 3, 4])
@pytest.mark.flaky(reruns=5)
@pytest.mark.integration_test
def test_power_law_converter_distribution(power):
    """
    Check that the distribution of resulting samples is uniform when
    converting from a power law.
    """
    c = PowerLawConverter(power, scale=1)
    x = stats.powerlaw(power + 1).rvs(size=10000)
    y, _ = c.to_uniform_parameter(x)
    d, p = stats.kstest(y, "uniform")
    assert p >= 0.05


@pytest.mark.parametrize("power", [1, 2, 3, 4])
@pytest.mark.integration_test
def test_power_law_converter_inversion(power):
    """
    Check that the power law inversion is invertible
    """
    c = PowerLawConverter(power, scale=1)
    x = stats.powerlaw(power + 1).rvs(size=1000)
    y, _ = c.to_uniform_parameter(x)
    x_out, _ = c.from_uniform_parameter(y)
    np.testing.assert_array_almost_equal(x, x_out)


@pytest.mark.requires("astropy")
@pytest.mark.parametrize("cosmology", ["Planck15", "WMAP7"])
@pytest.mark.integration_test
def test_comoving_distance_converter_integration(cosmology):
    """Integration test for the comoving distance prior converter."""
    cdc = ComovingDistanceConverter(
        d_min=10, d_max=1000, cosmology=cosmology, scale=100, n_interp=10
    )
    u, lju = cdc.to_uniform_parameter(500)
    dl, ljd = cdc.from_uniform_parameter(500)
    assert u is not None
    assert dl is not None
    assert lju == 0
    assert ljd == 0


@pytest.mark.requires("astropy")
@pytest.mark.parametrize("cosmology", ["Planck15", "WMAP7"])
@pytest.mark.parametrize("scale", [1, 1000])
@pytest.mark.parametrize("n_interp", [200, 500])
@pytest.mark.integration_test
def test_comoving_distance_converter_inversion(cosmology, scale, n_interp):
    """Integration test to verify the conversion is invertible."""
    cdc = ComovingDistanceConverter(
        d_min=100,
        d_max=5000,
        cosmology=cosmology,
        scale=scale,
        n_interp=n_interp,
    )
    dl = np.random.uniform(100, 5000, 100)
    u, _ = cdc.to_uniform_parameter(dl)
    dl_out, _ = cdc.from_uniform_parameter(u)
    np.testing.assert_array_almost_equal(dl, dl_out)



# -*- coding: utf-8 -*-
"""
Augmented version of FlowProposal.
"""
import logging

import numpy as np
from scipy import stats
from scipy.special import logsumexp

from .. import config
from .flowproposal import FlowProposal
from ..livepoint import numpy_array_to_live_points

logger = logging.getLogger(__name__)


class AugmentedFlowProposal(FlowProposal):
    """Version of FlowProposal that uses AugmentedFlows.

    Augmented normalising flows were proposed in: \
        https://arxiv.org/abs/2002.07101 and add auxiliary parameters to the
    inputs of the flow which are drawn from a Gaussian. This improves the
    flows' ability to learn multimodal distribution.

    Parameters
    ----------
    model : :obj:`nessai.model.Model`
        User defined model
    augment_dims : int
        Number of augment parameters to add to the inputs
    generate_augment : {'gaussian', 'zeroes', 'zeros'}, optional
        Method used when computing the radius of the latent contour.
    marginalise_augment : bool, optional
        Use the marginalised likelihood when performing rejection sampling.
        Adds significant computation cost.
    n_marg : int, optional
        Number of samples to use when approximating the marginalised
        likelihood.
    """

    def __init__(
        self,
        model,
        augment_dims=1,
        generate_augment="gaussian",
        marginalise_augment=False,
        n_marg=50,
        **kwargs,
    ):
        super().__init__(model, **kwargs)
        self.augment_dims = augment_dims
        self.generate_augment = generate_augment
        self.marginalise_augment = marginalise_augment
        self.n_marg = n_marg

    def set_rescaling(self):
        """Configure the rescaling.

        Calls the method from the parent class first and then adds the
        auxiliary parameters.
        """
        super().set_rescaling()
        # Cannot use super().rescale because rescale is changed in
        # set rescaling.
        self._base_rescale = self.rescale
        self.rescale = self._augmented_rescale
        self.augment_names = [f"e_{i}" for i in range(self.augment_dims)]
        self.names += self.augment_names
        self.rescaled_names += self.augment_names
        logger.info(f"augmented x space parameters: {self.names}")
        logger.info(f"parameters to rescale {self.rescale_parameters}")
        logger.info(
            f"Augmented x prime space parameters: {self.rescaled_names}"
        )

    def update_flow_config(self):
        """Update the flow configuration dictionary"""
        super().update_flow_config()
        m = np.ones(self.rescaled_dims)
        m[-self.augment_dims :] = -1
        if "kwargs" not in self.flow_config["model_config"].keys():
            self.flow_config["model_config"]["kwargs"] = {}
        self.flow_config["model_config"]["kwargs"]["mask"] = m

    def _augmented_rescale(
        self, x, generate_augment=None, compute_radius=False, **kwargs
    ):
        """Rescale with augment parameter.

        Parameters
        ----------
        x : array
            Structured array in X space with augment parameters in the
            fields.
        generate_augment : {None, 'gaussian', 'zeros', 'zeroes', False}, \
                optional
            Method used to generate the augmented parameters. If None Gaussian
            is used if compute_radius=False else attribute of the same name
            is used. If False no values are generated.
        compute_radius : bool, optional
            Boolean to indicate when rescale is being used for computing a
            radius.
        """
        x_prime, log_J = self._base_rescale(
            x, compute_radius=compute_radius, **kwargs
        )

        if generate_augment is None:
            if compute_radius:
                generate_augment = self.generate_augment
            else:
                generate_augment = "gaussian"

        if generate_augment in ["zeros", "zeroes"]:
            for an in self.augment_names:
                x_prime[an] = np.zeros(x_prime.size)
        elif generate_augment == "gaussian":
            for an in self.augment_names:
                x_prime[an] = np.random.randn(x_prime.size)
        else:
            raise RuntimeError("Unknown method for generating augment samples")

        return x_prime, log_J

    def augmented_prior(self, x):
        """
        Log Gaussian for augmented variables.

        If self.marginalise_augment is True, log_prior is 0.
        """
        log_p = 0.0
        if not self.marginalise_augment:
            for n in self.augment_names:
                log_p += stats.norm.logpdf(x[n])
        return log_p

    def log_prior(self, x):
        """
        Compute the prior probability in the non-prime space.
        """
        return super().log_prior(x) + self.augmented_prior(x)

    def x_prime_log_prior(self, x):
        """
        Compute prior probability in the prime space.
        """
        return super().x_prime_log_prior(x) + self.augmented_prior(x)

    def _marginalise_augment(self, x_prime):
        """Marginalise out the augmented features.

        Note that x_prime is not a structured array. See the original paper
        for the details
        """
        x_prime = np.repeat(x_prime, self.n_marg, axis=0)

        x_prime[:, -self.augment_dims :] = np.random.randn(
            x_prime.shape[0], self.augment_dims
        )

        _, log_prob = self.flow.forward_and_log_prob(x_prime)

        log_prob_e = np.sum(
            stats.norm.logpdf(x_prime[:, -self.augment_dims :]), axis=1
        )

        return -np.log(self.n_marg) + logsumexp(
            (log_prob - log_prob_e).reshape(-1, self.n_marg), axis=1
        )

    def backward_pass(self, z, rescale=True):
        """
        A backwards pass from the model (latent -> real)

        Parameters
        ----------
        z : array_like
            Structured array of points in the latent space
        rescale : bool, optional (True)
            Apply inverse rescaling function

        Returns
        -------
        x : array_like
            Samples in the latent space
        log_prob : array_like
            Log probabilities corresponding to each sample (including the
            Jacobian)
        """
        try:
            x, log_prob = self.flow.sample_and_log_prob(
                z=z, alt_dist=self.alt_dist
            )
        except AssertionError:
            return np.array([]), np.array([])

        if self.marginalise_augment:
            log_prob = self._marginalise_augment(x)

        valid = np.isfinite(log_prob)
        x, log_prob = x[valid], log_prob[valid]
        x = numpy_array_to_live_points(
            x.astype(config.livepoints.default_float_dtype),
            self.rescaled_names,
        )
        # Apply rescaling in rescale=True
        if rescale:
            x, log_J = self.inverse_rescale(x)
            # Include Jacobian for the rescaling
            log_prob -= log_J
            x, log_prob = self.check_prior_bounds(x, log_prob)
        return x, log_prob



"""Tests for livepoint functions"""

import sys

import numpy as np
import pandas as pd
import pytest

from nessai import config
import nessai.livepoint as lp
from nessai.utils.testing import assert_structured_arrays_equal


EXTRA_PARAMS_DTYPE = [
    (nsp, d)
    for nsp, d in zip(
        config.livepoints.non_sampling_parameters,
        config.livepoints.non_sampling_dtype,
    )
]


def assert_lp_equal(x, y, non_sampling_parameters=True):
    """Custom assertion that can skip the non-sampling parameters"""
    if non_sampling_parameters:
        assert_structured_arrays_equal(x, y)
    else:
        for nsp in config.livepoints.non_sampling_parameters:
            if nsp in x.dtype.names:
                raise AttributeError(
                    "x has non-sampling parameters but "
                    f"non_sampling_parameters=False. x dtype: {x.dtype}"
                )

        names = [
            n
            for n in x.dtype.names
            if n not in config.livepoints.non_sampling_parameters
        ]
        y = np.lib.recfunctions.repack_fields(y[names])
        assert_structured_arrays_equal(x, y)


@pytest.fixture(params=[True, False])
def non_sampling_parameters(request):
    return request.param


@pytest.fixture(autouse=True, params=[[], ["logQ", "logW"]])
def extra_parameters(request):
    """Add (and remove) extra parameters for the tests."""
    # Before every test
    lp.reset_extra_live_points_parameters()
    lp.add_extra_parameters_to_live_points(request.param)
    global EXTRA_PARAMS_DTYPE
    EXTRA_PARAMS_DTYPE = [
        (nsp, d)
        for nsp, d in zip(
            config.livepoints.non_sampling_parameters,
            config.livepoints.non_sampling_dtype,
        )
    ]

    # Test happens here
    yield

    # Called after every test
    lp.reset_extra_live_points_parameters()
    EXTRA_PARAMS_DTYPE = [
        (nsp, d)
        for nsp, d in zip(
            config.livepoints.non_sampling_parameters,
            config.livepoints.non_sampling_dtype,
        )
    ]


@pytest.fixture(params=["f4", "f16"])
def change_dtype(request):
    """Fixture that changes the default float dtype"""
    dtype = request.param
    if dtype == "f16" and sys.platform.startswith("win"):
        pytest.skip("Skipping test with float128 on Windows.")
    current_dtype = config.livepoints.default_float_dtype
    config.livepoints.default_float_dtype = dtype

    yield dtype

    config.livepoints.default_float_dtype = current_dtype


@pytest.fixture
def live_point():
    return np.array(
        [(1.0, 2.0, 3.0, *config.livepoints.non_sampling_defaults)],
        dtype=[
            ("x", config.livepoints.default_float_dtype),
            ("y", config.livepoints.default_float_dtype),
            ("z", config.livepoints.default_float_dtype),
        ]
        + EXTRA_PARAMS_DTYPE,
    )


@pytest.fixture
def live_points():
    return np.array(
        [
            (1.0, 2.0, 3.0, *config.livepoints.non_sampling_defaults),
            (4.0, 5.0, 6.0, *config.livepoints.non_sampling_defaults),
        ],
        dtype=[
            ("x", config.livepoints.default_float_dtype),
            ("y", config.livepoints.default_float_dtype),
            ("z", config.livepoints.default_float_dtype),
        ]
        + EXTRA_PARAMS_DTYPE,
    )


@pytest.fixture
def empty_live_point():
    return np.empty(
        0,
        dtype=[
            ("x", config.livepoints.default_float_dtype),
            ("y", config.livepoints.default_float_dtype),
            ("z", config.livepoints.default_float_dtype),
        ]
        + EXTRA_PARAMS_DTYPE,
    )


def test_add_extra_parameters(caplog):
    """Assert a warning is raised when adding a parameter name twice."""
    lp.add_extra_parameters_to_live_points(["test"], [1.0])
    assert "test" in config.livepoints.extra_parameters
    loc = config.livepoints.extra_parameters.index("test")
    assert config.livepoints.extra_parameters_defaults[loc] == 1.0
    lp.add_extra_parameters_to_live_points(["test"], [1.0])
    assert "Extra parameter `test` has already been added" in str(caplog.text)


def test_get_dtype():
    """Assert the correct value is returned"""
    names = ["x", "y"]
    expected = (
        [("x", "f4"), ("y", "f4")]
        + [
            ("logP", config.livepoints.default_float_dtype),
            ("logL", config.livepoints.logl_dtype),
            ("it", config.livepoints.it_dtype),
        ]
        + list(
            zip(
                config.livepoints.extra_parameters,
                config.livepoints.extra_parameters_dtype,
            )
        )
    )
    dtype = lp.get_dtype(names, array_dtype="f4")
    assert dtype == expected


def test_get_dtype_change_dtype(change_dtype):
    """Assert the dtype can be changed"""
    dtype = lp.get_dtype(["x"])
    assert dtype.fields["x"][0] == np.dtype(change_dtype)


def test_empty_structured_array_names(non_sampling_parameters):
    """Assert the correct default values are used when specifying names"""
    n = 10
    fields = ["x", "y"]
    array = lp.empty_structured_array(
        n, names=fields, non_sampling_parameters=non_sampling_parameters
    )
    for f in fields:
        np.testing.assert_array_equal(
            array[f], config.livepoints.default_float_value * np.ones(n)
        )
    for nsp, v in zip(
        config.livepoints.non_sampling_parameters,
        config.livepoints.non_sampling_defaults,
    ):
        if non_sampling_parameters:
            np.testing.assert_array_equal(array[nsp], v * np.ones(n))
        else:
            assert nsp not in array.dtype.names


def test_empty_structured_array_dtype(non_sampling_parameters):
    """Assert the correct default values are used when specifying the dtype"""
    n = 10
    dtype = [("x", "f8"), ("y", "f8")]
    if non_sampling_parameters:
        dtype += EXTRA_PARAMS_DTYPE
    array = lp.empty_structured_array(
        n, dtype=dtype, non_sampling_parameters=non_sampling_parameters
    )
    for f in ["x", "y"]:
        np.testing.assert_array_equal(
            array[f], config.livepoints.default_float_value * np.ones(n)
        )
    assert array.dtype == dtype


def test_empty_structured_array_zero_points(non_sampling_parameters):
    """Assert setting n=0 works correctly"""
    dtype = [("x", "f8"), ("y", "f8")]
    if non_sampling_parameters:
        dtype += EXTRA_PARAMS_DTYPE
    array = lp.empty_structured_array(
        0, names=["x", "y"], non_sampling_parameters=non_sampling_parameters
    )
    assert len(array) == 0
    assert array.dtype == dtype


def test_empty_structured_array_dtype_missing():
    """Assert an error is raised if the non-sampling parameters are missing"""
    n = 10
    dtype = [("x", "f8"), ("y", "f8")]
    with pytest.raises(ValueError) as excinfo:
        lp.empty_structured_array(n, dtype=dtype)
    assert "non-sampling parameters" in str(excinfo.value)


def test_empty_structured_array_change_dtype(change_dtype):
    """Assert changing the default dtype changes the output"""
    array = lp.empty_structured_array(1, names=["x"])
    assert array.dtype.fields["x"][0] == np.dtype(change_dtype)


def test_parameters_to_live_point(live_point, non_sampling_parameters):
    """
    Test function that produces a single live point given the parameter
    values for the live point as a live or an array
    """
    x = lp.parameters_to_live_point(
        [1.0, 2.0, 3.0],
        ["x", "y", "z"],
        non_sampling_parameters=non_sampling_parameters,
    )
    assert_lp_equal(
        x, live_point, non_sampling_parameters=non_sampling_parameters
    )


def test_empty_parameters_to_live_point(empty_live_point):
    """
    Test behaviour when an empty parameter is parsed
    """
    np.testing.assert_array_equal(
        lp.parameters_to_live_point([], ["x", "y", "z"]),
        empty_live_point,
    )


def test_numpy_array_to_live_point(live_point, non_sampling_parameters):
    """
    Test the function the produces an array of live points given numpy array
    of shape [# dimensions]
    """
    array = np.array([1.0, 2.0, 3.0])
    x = lp.numpy_array_to_live_points(
        array,
        names=["x", "y", "z"],
        non_sampling_parameters=non_sampling_parameters,
    )
    assert_lp_equal(
        x, live_point, non_sampling_parameters=non_sampling_parameters
    )


def test_numpy_array_multiple_to_live_points(
    live_points, non_sampling_parameters
):
    """
    Test the function the produces an array of live points given numpy array
    of shape [# point, # dimensions]
    """
    array = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])
    x = lp.numpy_array_to_live_points(
        array,
        names=["x", "y", "z"],
        non_sampling_parameters=non_sampling_parameters,
    )
    assert_lp_equal(
        x, live_points, non_sampling_parameters=non_sampling_parameters
    )


def test_empty_numpy_array_to_live_points(empty_live_point):
    """
    Test the function the produces an array of live points given an empty
    numpy array
    """
    np.testing.assert_array_equal(
        empty_live_point,
        lp.numpy_array_to_live_points(np.array([]), names=["x", "y", "z"]),
    )


@pytest.mark.parametrize(
    "d",
    [
        {"x": 1, "y": 2, "z": 3},
        {"x": 1.0, "y": 2.0, "z": 3.0},
    ],
)
def test_dict_to_live_point(live_point, d, non_sampling_parameters):
    """
    Test the function that converts a dictionary with a single live point to
    a live point array
    """
    x = lp.dict_to_live_points(
        d, non_sampling_parameters=non_sampling_parameters
    )
    assert_lp_equal(
        x, live_point, non_sampling_parameters=non_sampling_parameters
    )


@pytest.mark.parametrize(
    "d",
    [
        {"x": [1, 4], "y": [2, 5], "z": [3, 6]},
        {"x": np.array([1, 4]), "y": np.array([2, 5]), "z": np.array([3, 6])},
    ],
)
def test_dict_multiple_to_live_points(live_points, d, non_sampling_parameters):
    """
    Test the function that converts a dictionary with multiple live points to
    a live point array
    """
    x = lp.dict_to_live_points(
        d, non_sampling_parameters=non_sampling_parameters
    )
    assert_lp_equal(
        x, live_points, non_sampling_parameters=non_sampling_parameters
    )


def test_empty_dict_to_live_points(empty_live_point):
    """
    Test the function that converts a dictionary with empty lists to
    a live point array
    """
    np.testing.assert_array_equal(
        empty_live_point, lp.dict_to_live_points({"x": [], "y": [], "z": []})
    )


def test_dataframe_to_live_points(live_points, non_sampling_parameters):
    """Test converting from a pandas dataframe to live points."""
    df = pd.DataFrame({"x": [1, 4], "y": [2, 5], "z": [3, 6]})
    out = lp.dataframe_to_live_points(
        df, non_sampling_parameters=non_sampling_parameters
    )
    assert_lp_equal(out, live_points, non_sampling_parameters)


def test_live_point_to_numpy_array(live_point):
    """
    Test conversion from a live point to an unstructured numpy array
    """
    (
        np.array([[1, 2, 3, *config.livepoints.non_sampling_defaults]]),
        lp.live_points_to_array(live_point),
    )


def test_live_point_to_numpy_array_with_names(live_point):
    """
    Test conversion from a live point to an unstructured numpy array with
    specific fields
    """
    np.testing.assert_array_equal(
        np.array([[1, 3, np.nan]]),
        lp.live_points_to_array(live_point, names=["x", "z", "logP"]),
    )


def test_live_point_to_dict(live_point):
    """
    Test conversion of a live point to a dictionary
    """
    d = {"x": 1.0, "y": 2.0, "z": 3.0}
    d.update(
        {
            k: v
            for k, v in zip(
                config.livepoints.non_sampling_parameters,
                config.livepoints.non_sampling_defaults,
            )
        }
    )
    np.testing.assert_equal(lp.live_points_to_dict(live_point), d)


def test_live_point_to_dict_with_names(live_point):
    """
    Test conversion of a live point to a dictionary
    """
    d = {
        "x": np.array([1.0]),
        "z": np.array([3.0]),
        "logP": np.array([np.nan]),
    }
    np.testing.assert_equal(
        lp.live_points_to_dict(live_point, names=["x", "z", "logP"]), d
    )


def test_multiple_live_points_to_dict(live_points):
    """
    Test conversion of multiple_live points to a dictionary
    """
    d = {"x": [1, 4], "y": [2, 5], "z": [3, 6]}
    d.update(
        {
            k: 2 * [v]
            for k, v in zip(
                config.livepoints.non_sampling_parameters,
                config.livepoints.non_sampling_defaults,
            )
        }
    )
    d_out = lp.live_points_to_dict(live_points)
    assert list(d.keys()) == list(d_out.keys())
    np.testing.assert_array_equal(list(d.values()), list(d_out.values()))


def test_unstructured_view_dtype(live_points):
    """Assert the correct array is returned when given the dtype"""
    dtype = np.dtype({n: live_points.dtype.fields[n] for n in ["x", "y"]})
    view = lp.unstructured_view(live_points, dtype=dtype)
    assert view.base is live_points
    assert view.shape == (live_points.size, 2)


def test_unstructured_view_names(live_points):
    """Assert the correct array is returned when given names"""
    view = lp.unstructured_view(live_points, names=["x", "y"])
    assert view.base is live_points
    assert view.shape == (live_points.size, 2)


def test_unstructured_view_error(live_points):
    """Assert an error is raised when neither names or dtype is given."""
    with pytest.raises(TypeError):
        lp.unstructured_view(live_points)



# -*- coding: utf-8 -*-
"""Test the GW flow proposal method"""
from math import pi
import numpy as np
import pytest
from unittest.mock import create_autospec, MagicMock, patch

from nessai.gw.proposal import GWFlowProposal, AugmentedGWFlowProposal


@pytest.fixture
def proposal():
    return create_autospec(GWFlowProposal)


@pytest.fixture
def augmented_proposal():
    return create_autospec(AugmentedGWFlowProposal)


def test_get_reparameterisation(proposal):
    """Test to make sure the correct version of get reparameterisation is
    called.
    """
    with patch(
        "nessai.gw.proposal.get_gw_reparameterisation", return_value="reparam"
    ) as mock:
        out = GWFlowProposal.get_reparameterisation(proposal, "default")
    assert out == "reparam"
    mock.assert_called_once_with("default")


def test_add_default_reparameterisation(proposal):
    """Test the method `add_default_reparameterisation`"""
    proposal.aliases = GWFlowProposal.aliases
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.parameters = ["chirp_mass"]
    proposal.model = MagicMock()
    proposal.model.names = ["chirp_mass", "theta_jn"]
    proposal.model.bounds = {
        "chirp_mass": [10.0, 20.0],
        "theta_jn": [0.0, 3.0],
    }

    reparam = MagicMock()
    reparam.__name__ = "MockReparam"
    with patch(
        "nessai.gw.proposal.get_gw_reparameterisation",
        return_value=(reparam, {}),
    ) as mock_get:
        GWFlowProposal.add_default_reparameterisations(proposal)

    mock_get.assert_called_once_with("angle-sine")
    reparam.assert_called_once_with(
        parameters=["theta_jn"], prior_bounds={"theta_jn": [0.0, 3.0]}
    )


def test_add_default_reparameterisation_w_extra_params(proposal):
    """Test adding a reparameterisation that has extra parameters"""
    proposal.aliases = GWFlowProposal.aliases
    proposal.names = ["ra", "dec"]
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.parameters = []
    proposal.model = MagicMock()
    proposal.model.names = ["ra", "dec"]
    proposal.model.bounds = {
        "ra": [0.0, 2 * np.pi],
        "dec": [-np.pi / 2, np.pi / 2],
    }

    def add_parameters(*args):
        proposal._reparameterisation.parameters = ["ra", "dec"]

    proposal._reparameterisation.add_reparameterisation = MagicMock(
        side_effect=add_parameters
    )

    reparam = MagicMock()
    reparam.__name__ = "MockReparam"
    reparam.parameters = ["ra", "dec"]
    with patch(
        "nessai.gw.proposal.get_gw_reparameterisation",
        return_value=(reparam, {}),
    ) as mock_get:
        GWFlowProposal.add_default_reparameterisations(proposal)

    # Mustn't be called twice despite have two parameters.
    mock_get.assert_called_once_with("sky-ra-dec")
    reparam.assert_called_once_with(
        parameters=["ra", "dec"],
        prior_bounds={
            "ra": [0.0, 2.0 * np.pi],
            "dec": [-np.pi / 2, np.pi / 2],
        },
    )


def test_add_default_reparameterisation_unknown(proposal):
    """Test the method with an unknown parameters"""
    proposal.aliases = GWFlowProposal.aliases
    proposal.names = ["x"]
    proposal._reparameterisation = MagicMock()
    proposal._reparameterisation.parameters = []
    proposal.model = MagicMock()
    proposal.model.names = ["x"]
    proposal.model.bounds = {"x": [0, 10]}

    with patch(
        "nessai.gw.proposal.get_gw_reparameterisation",
        return_value=(None, {}),
    ) as mock_get:
        GWFlowProposal.add_default_reparameterisations(proposal)

    mock_get.assert_not_called()


def test_augmented_get_reparameterisation(augmented_proposal):
    """Test to make sure the correct version of get reparameterisation is
    called.
    """
    with patch(
        "nessai.gw.proposal.get_gw_reparameterisation", return_value="reparam"
    ) as mock:
        out = AugmentedGWFlowProposal.get_reparameterisation(
            augmented_proposal, "default"
        )
    assert out == "reparam"
    mock.assert_called_once_with("default")


def test_augmented_reparameterisation_prior(augmented_proposal):
    """Test to make sure the correct components of the log prior are being
    called.

    These are:
    - agumented_prior
    - FlowProposal.log_prior
    """
    augmented_proposal.augmented_prior = MagicMock()
    augmented_proposal._reparameterisation = MagicMock()
    augmented_proposal.model = MagicMock()

    with patch("nessai.proposal.flowproposal.FlowProposal.log_prior") as mock:
        AugmentedGWFlowProposal.log_prior(augmented_proposal, 1)

    augmented_proposal.augmented_prior.assert_called_once_with(1)
    mock.assert_called_once_with(1)


def test_augmented_reparameterisation_prime_prior(augmented_proposal):
    """Test to make sure the correct components of the primed_log prior are
    being called.

    These are:
    - agumented_prior
    - FlowProposal.x_prime_log_prior
    """
    augmented_proposal.augmented_prior = MagicMock()
    augmented_proposal._reparameterisation = MagicMock()

    with patch(
        "nessai.proposal.flowproposal.FlowProposal." "x_prime_log_prior"
    ) as mock:
        AugmentedGWFlowProposal.x_prime_log_prior(augmented_proposal, 1)

    augmented_proposal.augmented_prior.assert_called_once_with(1)
    mock.assert_called_once_with(1)


@pytest.mark.integration_test
def test_default_reparameterisations(caplog, tmpdir):
    """Assert that the GW defaults are used even in reparameterisations
    is not given.
    """
    caplog.set_level("INFO")
    model = MagicMock()
    model.names = ["mass_ratio", "theta_jn", "phase"]
    model.bounds = {
        "mass_ratio": [20.0, 40.0],
        "theta_jn": [0.0, pi],
        "phase": [0.0, 2 * pi],
    }
    model.reparameterisations = None
    expected_params = model.names + ["phase_radial"]
    proposal = GWFlowProposal(
        model, poolsize=100, output=str(tmpdir.mkdir("test"))
    )
    # Mocked model so can't verify rescaling
    proposal.verify_rescaling = MagicMock()
    proposal.initialise()
    assert proposal._reparameterisation is not None
    assert all(
        [p in expected_params for p in proposal._reparameterisation.parameters]
    )
    assert "reparameterisations included in GWFlowProposal" in str(caplog.text)



# -*- coding: utf-8 -*-
"""
Proposal method for initial sampling when priors can be sampled analytically.
"""
import datetime

import numpy as np

from .base import Proposal


class AnalyticProposal(Proposal):
    """Class for drawing samples from analytic priors.

    Will be used when ``nessai`` is called with ``analytic_priors=True`` and
    assumes the :py:meth:`nessai.model.Model.new_point` samples directly
    from the prior. This method must be implemented by the user.

    Parameters
    ----------
    args :
        Arguments passed to the parent class.
    poolsize : int, optional
        Number of points drawn at once.
    kwargs :
        Keyword arguments passed to the parent class.
    """

    def __init__(self, *args, poolsize=1000, **kwargs):
        super(AnalyticProposal, self).__init__(*args, **kwargs)
        self.populated = False
        self._poolsize = poolsize

    @property
    def poolsize(self):
        """Poolsize used for drawing new samples in batches."""
        return self._poolsize

    def populate(self, N=None):
        """
        Populate the pool by drawing from the priors.

        Will also evaluate the likelihoods if the proposal contains a
        multiprocessing pool.

        Parameters
        ----------
        N : int, optional
            Number of samples to draw. If not specified ``poolsize`` will be
            used.
        """
        if N is None:
            N = self.poolsize
        self.samples = self.model.new_point(N=N)
        self.samples["logP"] = self.model.batch_evaluate_log_prior(
            self.samples
        )
        self.indices = np.random.permutation(self.samples.shape[0]).tolist()
        self.samples["logL"] = self.model.batch_evaluate_log_likelihood(
            self.samples
        )
        self.populated = True

    def draw(self, old_sample, **kwargs):
        """
        Propose a new sample. Draws from the pool if it is populated, else
        it populates the pool.

        Parameters
        ----------
        old_sample : structured_array
            Old sample, this is not used in the proposal method
        kwargs :
            Keyword arguments passed to \
                :py:meth:`~nessai.proposal.analytic.AnalyticProposal.populate`
        """
        if not self.populated:
            st = datetime.datetime.now()
            self.populate(**kwargs)
            self.population_time += datetime.datetime.now() - st
        index = self.indices.pop()
        new_sample = self.samples[index]
        if not self.indices:
            self.populated = False
        return new_sample



# -*- coding: utf-8 -*-
"""
Utilities for sorting.
"""
from typing import List, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from ..reparameterisations import Reparameterisation


def sort_reparameterisations(
    reparameterisations: List["Reparameterisation"],
    existing_parameters: Optional[List[str]] = None,
    known_parameters: Optional[List[str]] = None,
    initial_sort: bool = True,
) -> List["Reparameterisation"]:
    """Sort reparameterisations based on their parameters and requirements.

    Parameters
    ----------
    reparameterisations : List[Reparameterisation]
        List of reparameterisations.
    existing_parameters : Optional[List[str]]
        List of parameters that are all included.
    known_parameters : Optional[List[str]]
        List of all known parameters. If not specified it is inferred from the
        list of reparameterisations.
    initial_sort : bool
        Toggle initial sorting by the number of requirements.

    Returns
    -------
    List[Reparameterisation]
        Sorted list of reparameterisations.

    Raises
    ------
    ValueError
        If a required parameter is missing from the known parameters.
    """
    ordered = []
    skipped = []
    parameters = existing_parameters.copy() if existing_parameters else []

    if known_parameters is None:
        known_parameters = parameters.copy()
        for r in reparameterisations:
            known_parameters += r.parameters

    if initial_sort:
        reparameterisations = sorted(
            reparameterisations, key=lambda r: len(r.requires)
        )

    for r in reparameterisations:
        if not r.requires or all([req in parameters for req in r.requires]):
            ordered.append(r)
            parameters += r.parameters
        elif any([req not in known_parameters for req in r.requires]):
            raise ValueError(
                f"{r.name} requires {r.requires} which contains parameters "
                f"that are not known ({known_parameters})"
            )
        else:
            skipped.append(r)

    if skipped:
        ordered += skipped
        return sort_reparameterisations(
            ordered,
            existing_parameters=existing_parameters,
            known_parameters=known_parameters,
            initial_sort=False,
        )
    return ordered



#!/usr/bin/env python

"""
Example of running nessai with bilby on a gravitational wave likelihood. This
examples includes all 15 parameters for CBC and should take around 2 hours to
run.

Based on the Bilby example: https://git.ligo.org/lscsoft/bilby
"""
import bilby
import numpy as np

outdir = "./outdir/"
label = "full_gw_example"

bilby.core.utils.setup_logger(outdir=outdir, label=label)

duration = 4.0
sampling_frequency = 2048.0

np.random.seed(151226)

# Use an injection that is similar to GW150914
injection_parameters = dict(
    total_mass=66.0,
    mass_ratio=0.9,
    a_1=0.4,
    a_2=0.3,
    tilt_1=0.5,
    tilt_2=1.0,
    phi_12=1.7,
    phi_jl=0.3,
    luminosity_distance=2000,
    theta_jn=0.4,
    psi=2.659,
    phase=1.3,
    geocent_time=1126259642.413,
    ra=1.375,
    dec=-1.2108,
)

waveform_arguments = dict(
    waveform_approximant="IMRPhenomPv2", reference_frequency=50.0
)

# Create the waveform_generator
waveform_generator = bilby.gw.waveform_generator.WaveformGenerator(
    sampling_frequency=sampling_frequency,
    duration=duration,
    frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
    parameter_conversion=(
        bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters
    ),
    waveform_arguments=waveform_arguments,
)

# Set up interferometers
ifos = bilby.gw.detector.InterferometerList(["H1", "L1", "V1"])
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"] - 3,
)
ifos.inject_signal(
    waveform_generator=waveform_generator, parameters=injection_parameters
)

# Set up prior
priors = bilby.gw.prior.BBHPriorDict()
priors["geocent_time"] = bilby.core.prior.Uniform(
    minimum=injection_parameters["geocent_time"] - 0.1,
    maximum=injection_parameters["geocent_time"] + 0.1,
    name="geocent_time",
    latex_label="$t_c$",
    unit="$s$",
)

# Initialise the likelihood
# nessai supports the marginalisation included in bilby
likelihood = bilby.gw.likelihood.GravitationalWaveTransient(
    interferometers=ifos,
    waveform_generator=waveform_generator,
    priors=priors,
    phase_marginalization=True,
    distance_marginalization=True,
)


# Run sampler

# The `flow_class` should be set to `GWFlowProposal` for GW PE. This includes
# specific default reparameterisations for certain parameters. For example,
# it knows that theta_jn is angle with a sine prior.

result = bilby.core.sampler.run_sampler(
    likelihood=likelihood,
    priors=priors,
    outdir=outdir,
    injection_parameters=injection_parameters,
    label=label,
    conversion_function=bilby.gw.conversion.generate_all_bbh_parameters,
    flow_class="GWFlowProposal",
    sampler="nessai",
    resume=False,
    plot=True,
    nlive=2000,
    maximum_uninformed=4000,
    seed=150914,
    analytic_priors=True,  # Bilby priors can be sampled from directly
    flow_config=dict(model_config=dict(n_blocks=6)),
    n_pool=2,
)

# Produce corner plots
result.plot_corner()



# -*- coding: utf-8 -*-
"""
Test the object that handles the nested sampling evidence and prior volumes.
"""
from unittest.mock import create_autospec, patch
import numpy as np
import pytest

from nessai.evidence import (
    _BaseNSIntegralState,
    _NSIntegralState,
    logsubexp,
)


@pytest.fixture()
def nlive():
    return 100


@pytest.fixture()
def base_state():
    return create_autospec(_BaseNSIntegralState)


@pytest.fixture()
def ns_state():
    return create_autospec(_NSIntegralState)


def test_logsubexp_negative():
    """
    Test behaviour of logsubexp for x < y
    """
    with pytest.raises(Exception):
        logsubexp(1, 2)


@pytest.mark.parametrize(
    "method", ["log_evidence", "log_evidence_error", "log_posterior_weights"]
)
def test_base_state_not_implemented_error_properties(base_state, method):
    """Assert all of the abstract properties raised not implemented errors"""
    with pytest.raises(NotImplementedError):
        getattr(_BaseNSIntegralState, method).__get__(base_state)


@pytest.mark.parametrize(
    "weights, expected", [([], 0), ([0.1, 0.2, 0.3, 0.4], (1 / 0.3))]
)
def test_base_state_effective_n_posterior_samples(
    base_state, weights, expected
):
    """Assert the correct effective sample size is returned.

    Should be zero if weights are empty.
    """
    base_state.log_posterior_weights = np.log(weights)
    ess = _BaseNSIntegralState.effective_n_posterior_samples.__get__(
        base_state
    )
    np.testing.assert_almost_equal(ess, expected, decimal=10)


def test_invalid_expectation(ns_state):
    """Assert an error is raised if `expectation` is an invalid value"""
    with pytest.raises(
        ValueError, match=r"Expectation must be t or logt, got: a"
    ):
        _NSIntegralState.__init__(ns_state, 100, expectation="a")


@pytest.mark.parametrize("expectation", ["logt", "t"])
def test_increment(nlive, expectation):
    """Test the basic functionality of incrementing the evidence estimate"""
    state = _NSIntegralState(nlive, expectation=expectation)
    state.increment(-10)

    if expectation == "logt":
        target = -1 / nlive
    else:
        target = -np.log1p(1 / nlive)

    assert state.logw == target
    assert state.logZ != -np.inf
    np.testing.assert_equal(state.logLs, [-np.inf, -10])


def test_increment_monotonic_warning(ns_state, caplog):
    """Assert a warning is raised if the likelihood is non-monotonic"""
    ns_state.logLs = [1, 2, 3]
    ns_state.base_nlive = 10
    ns_state.nlive = 3 * [10]
    ns_state.logZ = -2.0
    ns_state.logw = 1.0
    ns_state.info = [0]
    ns_state.log_vols = []
    ns_state.track_gradients = False
    ns_state.expectation = "logt"
    _NSIntegralState.increment(ns_state, 2.5)
    assert "received non-monotonic logL" in str(caplog.text)


def test_log_evidence(ns_state):
    """Assert the log-evidence property returns the correct value"""
    expected = 1.0
    ns_state.logZ = expected
    out = _NSIntegralState.log_evidence.__get__(ns_state)
    assert out == expected


def test_log_evidence_error(ns_state, nlive):
    """Assert the log-evidence error property returns the correct value"""
    expected = np.sqrt(10 / nlive)
    ns_state.info = [1, 5, 10]
    ns_state.base_nlive = nlive
    out = _NSIntegralState.log_evidence_error.__get__(ns_state)
    assert out == expected


@pytest.mark.parametrize("nlive", [10, 20])
@pytest.mark.parametrize("expectation", ["t", "logt"])
def test_get_logx_live_points(ns_state, nlive, expectation):
    nit = 100
    nlive_per_iteration = np.arange(nlive, 0, -1, dtype=float)

    ns_state.expectation = expectation

    if expectation == "logt":
        logt = -1.0 / nlive
        logt_final = -1.0 / nlive_per_iteration
    elif expectation == "t":
        logt = -np.log1p(1.0 / nlive)
        logt_final = -np.log1p(1.0 / nlive_per_iteration)
    else:
        raise ValueError

    ns_state.logw = nit * logt
    expected = nit * logt + np.cumsum(logt_final)
    out = _NSIntegralState.get_logx_live_points(ns_state, nlive)
    np.testing.assert_equal(out, expected)


def test_finalise(nlive):
    """Test the check that finalise returns an improved logZ"""
    state = _NSIntegralState(nlive)
    state.increment(-10)
    pre = state.logZ
    state.finalise()
    assert state.logZ != -np.inf
    assert pre != state.logZ


def test_info(nlive):
    """Test to check the information increases as expected"""
    state = _NSIntegralState(nlive)
    state.increment(-10)
    assert state.info == [0.0]
    state.increment(-5)
    assert state.info[1] > 0


def test_track_gradients(nlive):
    """
    Test to make sure gradients are not computed when tracking is disabled
    """
    state = _NSIntegralState(nlive, track_gradients=False)
    state.increment(-10)
    state.increment(-5)
    assert len(state.gradients) == 1


@pytest.mark.parametrize(
    "expectation, value", [("logt", -1 / 50), ("t", -np.log1p(1 / 50))]
)
def test_variable_nlive(nlive, expectation, value):
    """
    Test to make sure that the using a different nlive changes the update
    values.
    """
    state = _NSIntegralState(nlive, expectation=expectation)
    state.increment(-10, nlive=50)
    assert state.logw == value


def test_plot(nlive):
    """Test the plotting function"""
    state = _NSIntegralState(nlive)
    state.increment(-10)
    state.increment(-5)
    fig = state.plot()
    assert fig is not None


def test_plot_w_filename(nlive, tmpdir):
    """Test the plotting function with a filename specified"""
    filename = str(tmpdir.mkdir("test"))
    state = _NSIntegralState(nlive)
    state.increment(-10)
    state.increment(-5)
    fig = state.plot(filename=filename)
    assert fig is None


def test_log_posterior_weights(ns_state):
    """Test the log-posterior weights property"""
    log_vols = [0.0, -0.1, -0.2, -0.3]
    logL = [np.NINF, -10.0, -5.0, -0.0]
    ns_state.log_vols = log_vols
    ns_state.logLs = logL
    log_z = -1.0
    with patch(
        "nessai.evidence.log_integrate_log_trap", return_value=log_z
    ) as mock_int:
        out = _NSIntegralState.log_posterior_weights.__get__(ns_state)

    trap_inputs = mock_int.call_args[0]
    np.testing.assert_array_equal(trap_inputs[0], logL + [-0.0])
    np.testing.assert_array_equal(trap_inputs[1], log_vols + [np.NINF])
    # Output should one shorted than logL since the initial point is not a
    # nested sample.
    assert len(out) == (len(logL) - 1)



import pytest
from unittest.mock import create_autospec

from nessai.samplers.nestedsampler import NestedSampler


@pytest.fixture(scope="function")
def sampler(model):
    s = create_autospec(NestedSampler)
    s.nlive = 100
    s.model = model
    s.store_live_points = False
    return s



from nessai.samplers.importancesampler import (
    ImportanceNestedSampler as INS,
    OrderedSamples,
)
from nessai.evidence import _INSIntegralState
from unittest.mock import MagicMock


def test_log_posterior_weights_property(ins):
    ins._ordered_samples = MagicMock(spec=OrderedSamples)
    ins._ordered_samples.state = MagicMock(spec=_INSIntegralState)
    assert (
        INS.log_posterior_weights.__get__(ins)
        is ins._ordered_samples.state.log_posterior_weights
    )



# -*- coding: utf-8 -*-
"""
Utilities related to rescaling.
"""
import logging

import numpy as np

from .hist import auto_bins

logger = logging.getLogger(__name__)


def rescale_zero_to_one(x, xmin, xmax):
    """
    Rescale a value to 0 to 1

    Parameters
    ----------
    x : ndarray
        Array of values to rescale
    xmin, xmax : floats
        Minimum and maximum values to use for rescaling

    Returns
    -------
    ndarray
        Array of rescaled values
    ndarray
        Array of log determinants of Jacobians for each sample
    """
    return (x - xmin) / (xmax - xmin), -np.log(xmax - xmin)


def inverse_rescale_zero_to_one(x, xmin, xmax):
    """
    Rescale from 0 to 1 to xmin to xmax

    Parameters
    ----------
    x : ndarray
        Array of values to rescale
    xmin, xmax : floats
        Minimum and maximum values to use for rescaling

    Returns
    -------
    ndarray
        Array of rescaled values
    ndarray
        Array of log determinants of Jacobians for each sample
    """
    return (xmax - xmin) * x + xmin, np.log(xmax - xmin)


def rescale_minus_one_to_one(x, xmin, xmax):
    """
    Rescale a value to -1 to 1

    Parameters
    ----------
    x : ndarray
        Array of values to rescale
    xmin, xmax : floats
        Minimum and maximum values to use for rescaling

    Returns
    -------
    ndarray
        Array of rescaled values
    ndarray
        Array of log determinants of Jacobians for each sample
    """
    return (
        (2.0 * (x - xmin) / (xmax - xmin)) - 1,
        np.log(2) - np.log(xmax - xmin),
    )


def inverse_rescale_minus_one_to_one(x, xmin, xmax):
    """
    Rescale from -1 to 1 to xmin to xmax

    Parameters
    ----------
    x : ndarray
        Array of values to rescale
    xmin, xmax : floats
        Minimum and maximum values to use for rescaling

    Returns
    -------
    ndarray
        Array of rescaled values
    ndarray
        Array of log determinants of Jacobians for each sample
    """
    return (
        (xmax - xmin) * ((x + 1) / 2.0) + xmin,
        np.log(xmax - xmin) - np.log(2),
    )


def detect_edge(
    x,
    x_range=None,
    percent=0.1,
    cutoff=0.5,
    nbins="auto",
    allow_both=False,
    allow_none=False,
    allowed_bounds=["lower", "upper"],
    test=None,
):
    """
    Detect edges in input distributions based on the density.

    Parameters
    ----------
    x: array_like
        Samples
    x_range : array_like, optional
        Lower and upper bounds used to check inversion, if not specified
        min and max of data are used.
    percent: float (0.1)
        Percentage of interval used to check edges
    cutoff: float (0.1)
        Minimum fraction of the maximum density contained within the
        percentage of the interval specified
    nbins : float or 'auto'
        Number of bins used for histogram.
    allow_both: bool
        Allow function to return both instead of force either upper or lower
    allow_none: bool
        Allow for neither lower or upper bound to be returned
    allowed_bounds : list
        List of alloweds bounds.
    test : str or None
        If not None this skips the process and just returns the value of test.
        This is used to verify the inversion in all possible scenarios.

    Returns
    -------
    str or False, {'lower', 'upper', 'both', False}
        Returns the boundary to apply the inversion or False is no inversion
        is to be applied
    """
    bounds = ["lower", "upper"]
    if test is not None:
        logger.debug("Using test in detect_edge")
        if test in bounds and test not in allowed_bounds:
            logger.debug(f"{test} is not an allowed bound, returning False")
            return False
        else:
            return test
    if not all(b in bounds for b in allowed_bounds):
        raise RuntimeError(f"Unknown allowed bounds: {allowed_bounds}")
    if nbins == "auto":
        nbins = auto_bins(x)

    hist, bins = np.histogram(x, bins=nbins, density=True, range=x_range)
    n = max(int(len(bins) * percent), 1)
    bounds_fraction = np.array([np.sum(hist[:n]), np.sum(hist[-n:])]) * (
        bins[1] - bins[0]
    )
    max_idx = np.argmax(hist)
    max_density = hist[max_idx] * (bins[1] - bins[0])
    logger.debug(f"Max. density: {max_density:.3f}")

    for i, b in enumerate(bounds):
        if b not in allowed_bounds:
            bounds.pop(i)
            bounds_fraction = np.delete(bounds_fraction, i)

    if (
        np.all(bounds_fraction > cutoff * max_density)
        and allow_both
        and len(bounds) > 1
    ):
        logger.debug("Both bounds above cutoff")
        return "both"
    elif max_idx < n and "lower" in bounds:
        return bounds[0]
    elif max_idx >= (len(bins) - n) and "upper" in bounds:
        return bounds[-1]
    elif not np.any(bounds_fraction > cutoff * max_density) and allow_none:
        logger.debug("Density too low at both bounds")
        return False
    else:
        logger.debug("No bound preferred, returning bound with higher density")
        return bounds[np.argmax(bounds_fraction)]


def configure_edge_detection(d, detect_edges):
    """
    Configure parameters for edge detection

    Parameters
    ----------
    d : dict
        Dictionary of kwargs passed to detect_edge.
    detect_edges : bool
        If true allows for no inversion to be applied.

    Returns
    -------
    dict
        Updated kwargs
    """
    default = dict(cutoff=0.5)
    if d is None:
        d = {}
    if detect_edges:
        d["allow_none"] = True
    else:
        d["allow_none"] = False
        d["cutoff"] = 0.0
    default.update(d)
    logger.debug(f"detect edges kwargs: {default}")
    return default


def determine_rescaled_bounds(
    prior_min,
    prior_max,
    x_min,
    x_max,
    invert=None,
    inversion=False,
    offset=0,
    rescale_bounds=[-1, 1],
):
    """
    Determine the values of the prior min and max in the rescaled
    space.

    Parameters
    ----------
    prior_min : float
        Minimum of the prior.
    prior_max : float
        Maximum of the prior.
    x_min : float
        New minimum.
    x_max : float
        New maximum.
    invert : False or {'upper', 'lower', 'both'}, optional
        Type of inversion. `inversion` must also be set to True.
    inversion : bool, optional
        Indicate if the rescaling bounds have been set for inversion. If True
        and invert is None or False, then the rescale bounds are assumed to
        be [-1, 1] rather than [0, 1] (the default for inverted parameters.)
    offset : float, optional
        Offset to subtract from the values prior to rescaling.
    rescaled_bounds : list or tuple
        Lower and upper bound which x has been rescaled to. In inversion=True,
        these values are ignored to match behaviour in the \
            :py:class:`~nessai.reparameterisations.RescaleToBounds`.
    """
    if x_min == x_max:
        raise ValueError("New minimum and maximum are equal")
    if not inversion:
        scale = rescale_bounds[1] - rescale_bounds[0]
        shift = rescale_bounds[0]
    else:
        scale = 1.0
        shift = 0.0
    lower = scale * (prior_min - offset - x_min) / (x_max - x_min) + shift
    upper = scale * (prior_max - offset - x_min) / (x_max - x_min) + shift
    if not inversion:
        if invert:
            logger.warning(
                "`invert` is not False or None, but `inversion=False`"
            )
        return lower, upper
    elif not invert or invert is None:
        return 2 * lower - 1, 2 * upper - 1
    elif invert == "upper":
        return lower - 1, 1 - lower
    elif invert == "lower":
        return -upper, upper
    elif invert == "both":
        return -0.5, 1.5
    else:
        raise ValueError(f"Invalid value for `invert`: {invert}")


def logit(x, eps=None):
    """Logit function that also returns log Jacobian determinant.

    See :py:func:`nessai.utils.rescaling.sigmoid` for the inverse.

    Parameters
    ----------
    x : float or ndarray
        Array of values
    eps : float, optional
        Epsilon value used to clamp inputs to [eps, 1 - eps]. If None, then
        inputs are not clamped.

    Returns
    -------
    float or ndarray
        Rescaled values.
    float or ndarray
        Log Jacobian determinant.
    """
    if eps:
        x = np.clip(x, eps, 1 - eps)
    log_j = -np.log(x) - np.log1p(-x)
    return np.log(x) - np.log1p(-x), log_j


def sigmoid(x):
    """Sigmoid function that also returns log Jacobian determinant.

    See :py:func:`nessai.utils.rescaling.logit` for the inverse.

    Parameters
    ----------
    x : float or ndarray
        Array of values

    Returns
    -------
    float or ndarray
        Rescaled values.
    float or ndarray
        Log Jacobian determinant.
    """
    with np.errstate(over="ignore", divide="ignore"):
        x = np.divide(1, 1 + np.exp(-x))
        log_j = np.log(x) + np.log1p(-x)
    return x, log_j


def logistic_function(x, x0=0.0, k=1.0):
    """Logistic function with configurable midpoint and gradient.

    Parameters
    ----------
    x : np.ndarray
        Samples to apply function to
    x0 : float
        Midpoint
    k: float
        Gradient

    Returns
    -------
    np.ndarrary
        Value of logistic function for each x.
    """
    return 1.0 / (1 + np.exp(-k * (x - x0)))


def log_with_log_jacobian(x):
    """Natural log and log-Jacobian determinant.

    Parameters
    ----------
    x : Union[float, np.ndarray]
        Array of values

    Returns
    -------
    Union[float, numpy.ndarray]
        Rescaled values.
    Union[float, numpy.ndarray]
        Log Jacobian determinant.
    """
    y = np.log(x)
    return y, -y


def exp_with_log_jacobian(x):
    """Exponential and log-Jacobian determinant.

    Parameters
    ----------
    x : Union[float, np.ndarray]
        Array of values

    Returns
    -------
    Union[float, numpy.ndarray]
        Rescaled values.
    Union[float, numpy.ndarray]
        Log Jacobian determinant.
    """
    return np.exp(x), x.copy()


rescaling_functions = {
    "logit": (logit, sigmoid),
    "log": (log_with_log_jacobian, exp_with_log_jacobian),
    "exp": (exp_with_log_jacobian, log_with_log_jacobian),
}



# -*- coding: utf-8 -*-
"""Test the flow utilities"""
import logging
import numpy as np
import pytest
import torch
import torch.nn.functional as F
from unittest.mock import MagicMock, patch

from nessai.flows.utils import (
    configure_model,
    create_linear_transform,
    create_pre_transform,
    get_base_distribution,
    get_n_neurons,
    silu,
    reset_weights,
    reset_permutations,
)


@pytest.fixture
def config():
    """Minimal config needed for configure_model to work"""
    return dict(
        n_inputs=2, n_neurons=4, n_blocks=2, n_layers=1, ftype="realnvp"
    )


def test_silu():
    """Test the silu activation"""
    from scipy.special import expit

    x = torch.randn(100)
    y = silu(x)
    expected = x.numpy() * expit(x.numpy())
    np.testing.assert_array_almost_equal(y, expected)


def test_get_base_distribution_none():
    """Assert that if the distribution is None, None is returned"""
    assert get_base_distribution(2, None) is None


def test_get_base_distribution_class_instance():
    """Assert that if the distribution is a class instance it is returned"""
    dist = MagicMock()
    assert get_base_distribution(2, dist) is dist


def test_get_base_distribution_class():
    """Assert that if the distribution is a class is it correctly called"""
    from nessai.flows.distributions import MultivariateNormal

    dist_cls = MultivariateNormal
    dist = get_base_distribution(2, dist_cls, var=2)
    assert isinstance(dist, MultivariateNormal)
    assert dist._var == 2


def test_get_base_distribution_str():
    """Assert that if the distribution is a str it is correctly called"""
    from nessai.flows.distributions import MultivariateNormal

    dist = get_base_distribution(2, "mvn", var=2)
    assert isinstance(dist, MultivariateNormal)
    assert dist._var == 2


def test_get_base_distribution_error():
    """Assert an unknown dist raises an error."""
    with pytest.raises(ValueError) as excinfo:
        get_base_distribution(2, "not_a_distribution")
    assert "Unknown distribution: not_a_distribution" in str(excinfo.value)


@pytest.mark.parametrize("name", ["lars", "resampled"])
@pytest.mark.parametrize("n_layers", [None, 3])
def test_get_base_distribution_lars(name, n_layers):
    """Test using the LARS base distribution"""
    kwargs = dict(test=False, net_kwargs=dict(dropout=0.1))
    if n_layers:
        kwargs["n_layers"] = n_layers
    else:
        n_layers = 2

    mlp = object()

    with patch("nessai.flows.utils.ResampledGaussian") as mock_dist, patch(
        "nessai.flows.utils.MLP", return_value=mlp
    ) as mock_mlp, patch(
        "nessai.flows.utils.get_n_neurons", return_value=8
    ) as mock_get_neurons:
        get_base_distribution(2, name, **kwargs)

    mock_get_neurons.assert_called_once_with(None, n_inputs=2)
    mock_mlp.assert_called_once_with(
        [2], [1], n_layers * [8], activate_output=torch.sigmoid, dropout=0.1
    )
    mock_dist.assert_called_once_with([2], mlp, test=False)


@pytest.mark.parametrize(
    "n_neurons, n_inputs, expected",
    [
        (16, 2, 16),
        ("auto", 2, 4),
        ("double", 2, 4),
        (None, 2, 4),
        ("equal", 2, 2),
        ("half", 4, 2),
        (None, None, 8),
    ],
)
def test_get_n_neurons(n_neurons, n_inputs, expected):
    """Assert the correct values are returned"""
    out = get_n_neurons(n_neurons=n_neurons, n_inputs=n_inputs)
    assert isinstance(out, int)
    assert out == expected


@pytest.mark.parametrize(
    "n_neurons, n_inputs",
    [
        ("auto", None),
        ("half", None),
        ("equal", None),
        ("double", None),
        ("invalid", 4),
    ],
)
def test_get_n_neurons_value_error(n_neurons, n_inputs):
    """Assert a ValueError is raised if invalid inputs are given"""
    with pytest.raises(ValueError) as excinfo:
        get_n_neurons(n_neurons=n_neurons, n_inputs=n_inputs)
    assert "Could not get number of neurons" in str(excinfo.value)


def test_reset_weights_with_reset_parameters():
    """Test the reset weights function for module with ``reset_parameters``"""
    module = MagicMock()
    module.reset_parameters = MagicMock()
    reset_weights(module)
    module.reset_parameters.assert_called_once()


def test_reset_weights_batch_norm():
    """Test the reset weights function for an instance of batch norm"""
    from glasflow.nflows.transforms.normalization import BatchNorm

    x = torch.randn(20, 2)
    module = BatchNorm(2, eps=0.1)
    module.train()
    module.forward(x)

    reset_weights(module)

    constant = np.log(np.exp(1 - 0.1) - 1)
    assert (module.unconstrained_weight.data == constant).all()
    assert (module.bias.data == 0).all()
    assert (module.running_mean == 0).all()
    assert (module.running_var == 1).all()


def test_reset_weights_other_module(caplog):
    """Test reset weights on a module that cannot be reset."""
    caplog.set_level(logging.WARNING)
    module = object
    reset_weights(module)
    assert "Could not reset" in caplog.text


def test_weight_reset_permutation():
    """Test to make sure random permutation is reset correctly"""
    from glasflow.nflows.transforms.permutations import RandomPermutation

    x = torch.arange(10).reshape(1, -1)
    m = RandomPermutation(features=10)
    y_init, _ = m(x)
    p = m._permutation.numpy()
    m.apply(reset_permutations)
    y_reset, _ = m(x)
    assert not (p == m._permutation.numpy()).all()
    assert not (y_init.numpy() == y_reset.numpy()).all()


def test_reset_permutation_lu():
    """Assert LULinear is reset correctly"""
    from glasflow.nflows.transforms import LULinear
    from glasflow.nflows.transforms.linear import LinearCache

    lu = MagicMock(spec=LULinear)
    lu.cache = MagicMock(spec=LinearCache)
    reset_permutations(lu)
    lu.cache.invalidate.assert_called_once()
    lu._initialize.assert_called_once_with(identity_init=True)


def test_configure_model_basic(config):
    """Test configure model with the most basic config."""
    config["kwargs"] = dict(num_bins=2)
    with patch("nessai.flows.realnvp.RealNVP") as mock_flow:
        configure_model(config)

    mock_flow.assert_called_with(
        config["n_inputs"],
        config["n_neurons"],
        config["n_blocks"],
        config["n_layers"],
        num_bins=2,
    )


@pytest.mark.parametrize(
    "flow_inputs",
    [
        {"ftype": "realnvp", "expected": "realnvp.RealNVP"},
        {"ftype": "frealnvp", "expected": "realnvp.RealNVP"},
        {"ftype": "spline", "expected": "nsf.NeuralSplineFlow"},
        {"ftype": "nsf", "expected": "nsf.NeuralSplineFlow"},
        {"ftype": "maf", "expected": "maf.MaskedAutoregressiveFlow"},
    ],
)
def test_configure_model_flows(config, flow_inputs):
    """Test the different flows."""
    config["ftype"] = flow_inputs["ftype"]
    with patch(f"nessai.flows.{flow_inputs['expected']}") as mock_flow:
        model, _ = configure_model(config)
    mock_flow.assert_called_with(
        config["n_inputs"],
        config["n_neurons"],
        config["n_blocks"],
        config["n_layers"],
    )


def test_configure_model_flow_class(config):
    """Test using a custom class of flow."""

    class TestFlow:
        def __init__(self, n_inputs, n_neurons, n_blocks, n_layers):
            self.n_inputs = n_inputs
            self.n_neurons = n_neurons
            self.n_blocks = n_blocks
            self.n_layers = n_layers

        def to(self, input):
            pass

    config["flow"] = TestFlow
    model, _ = configure_model(config)
    assert isinstance(model, TestFlow)
    assert model.n_inputs == config["n_inputs"]
    assert model.n_neurons == config["n_neurons"]
    assert model.n_blocks == config["n_blocks"]
    assert model.n_layers == config["n_layers"]


def test_configure_model_device_cuda(config):
    config["device_tag"] = "cuda"
    expected_device = torch.device("cuda")
    mock_model = MagicMock()
    with patch(
        "nessai.flows.realnvp.RealNVP", return_value=mock_model
    ) as mock_flow:
        model, device = configure_model(config)

    mock_flow.assert_called_with(
        config["n_inputs"],
        config["n_neurons"],
        config["n_blocks"],
        config["n_layers"],
    )

    mock_model.to.assert_called_once_with(expected_device)
    assert model.device == expected_device
    assert device == expected_device


@pytest.mark.parametrize(
    "act",
    [
        {"act": "relu", "expected": F.relu},
        {"act": "tanh", "expected": F.tanh},
        {"act": "silu", "expected": silu},
        {"act": "swish", "expected": silu},
    ],
)
def test_configure_model_activation_functions(config, act):
    """Test the different activation functions."""
    config["kwargs"] = dict(activation=act["act"])

    with patch("nessai.flows.realnvp.RealNVP") as mock_flow:
        configure_model(config)

    mock_flow.assert_called_with(
        config["n_inputs"],
        config["n_neurons"],
        config["n_blocks"],
        config["n_layers"],
        activation=act["expected"],
    )


def test_configure_model_distribution(config):
    """Assert distribution is added to the kwargs"""
    config["distribution"] = "mvn"
    dist = MagicMock()
    with patch(
        "nessai.flows.utils.get_base_distribution", return_value=dist
    ), patch("nessai.flows.realnvp.RealNVP") as mock:
        configure_model(config)
    assert "distribution" in mock.call_args[1]
    assert mock.call_args[1]["distribution"] is dist


def test_configure_model_ftype_error(config):
    """Assert unknown types of flow raise an error."""
    config.pop("ftype")
    with pytest.raises(RuntimeError) as excinfo:
        configure_model(config)
    assert "Must specify either 'flow' or 'ftype'." in str(excinfo.value)


def test_configure_model_input_type_error(config):
    """Assert incorrect type for n_inputs raises an error."""
    config["n_inputs"] = "10"
    with pytest.raises(TypeError) as excinfo:
        configure_model(config)
    assert "Number of inputs (n_inputs) must be an int" in str(excinfo.value)


def test_configure_model_unknown_activation(config):
    """Assert unknown activation functions raise an error"""
    config["kwargs"] = dict(activation="test")
    with pytest.raises(RuntimeError) as excinfo:
        configure_model(config)
    assert "Unknown activation function: 'test'" in str(excinfo.value)


def test_configure_model_invalid_device(caplog, config):
    """Assert warning is raised and the device is set to CPU"""
    from nessai.flows.base import NFlow

    config["device_tag"] = "cpu"
    flow = MagicMock(spec=NFlow)

    def raise_error(input):
        raise RuntimeError("An error")

    flow.to = MagicMock(side_effect=raise_error)
    with caplog.at_level(logging.WARNING), patch(
        "nessai.flows.realnvp.RealNVP", return_value=flow
    ):
        configure_model(config)
    assert "Could not send the normalising flow to" in caplog.text


def test_configure_model_invalid_key(config):
    """Assert an error is raised if invalid keys are present in the config"""
    config["invalid_key"] = True
    with pytest.raises(
        RuntimeError, match=r"Unknown keys in model config: {'invalid_key'}"
    ):
        configure_model(config)


@pytest.mark.parametrize("linear_transform", ["lu", "permutation", "svd"])
def test_create_linear_transform(linear_transform):
    """Test creating a linear transform."""
    lt = create_linear_transform(linear_transform, 2)
    assert lt is not None


def test_create_linear_transform_unknown():
    """Assert an error is raised if an invalid input is given."""
    with pytest.raises(ValueError) as excinfo:
        create_linear_transform("not_a_transform", 2)
    assert "Unknown linear transform: not_a_transform" in str(excinfo.value)


@pytest.mark.parametrize("pre_transform", ["logit", "batch_norm"])
def test_create_pre_transform(pre_transform):
    """Test creating a pre-transform"""
    out = create_pre_transform(pre_transform, 2)
    assert out is not None


def test_create_pre_transform_unknown():
    """Assert an error is raised for an unknown pre-transform"""
    with pytest.raises(ValueError) as excinfo:
        create_pre_transform("not_a_transform", 2)
    assert "Unknown pre-transform: not_a_transform" in str(excinfo.value)
